# General imports
from dataclasses import dataclass
from pathlib import Path

# ML imports
import matplotlib.pyplot as plt
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from pytorch_lightning.loggers import WandbLogger
from pytorch_msssim import ms_ssim
from torch.utils import data

from power_perceiver.analysis.plot_satellite import LogSatellitePlots

# power_perceiver imports
from power_perceiver.consts import NUM_HIST_SAT_IMAGES, BatchKey
from power_perceiver.load_prepared_batches.data_sources import HRVSatellite
from power_perceiver.load_prepared_batches.data_sources.satellite import SAT_MEAN, SAT_STD
from power_perceiver.load_prepared_batches.prepared_dataset import PreparedDataset
from power_perceiver.pytorch_modules.satellite_predictor import XResUNet

plt.rcParams["figure.figsize"] = (18, 10)
plt.rcParams["figure.facecolor"] = "white"

DATA_PATH = Path("/home/jack/data/v15")
#  "/mnt/storage_ssd_4tb/data/ocf/solar_pv_nowcasting/nowcasting_dataset_pipeline/"
#  "prepared_ML_training_data/v15/"
assert DATA_PATH.exists()
NUM_FUTURE_SAT_IMAGES = 24


def get_dataloader(data_path: Path, tag: str) -> data.DataLoader:
    assert tag in ["train", "validation"]
    assert data_path.exists()

    dataset = PreparedDataset(
        data_path=data_path,
        data_loaders=[
            HRVSatellite(),
        ],
    )

    dataloader = data.DataLoader(
        dataset,
        batch_size=None,
        num_workers=12,
        pin_memory=True,
    )

    return dataloader


train_dataloader = get_dataloader(DATA_PATH / "train", tag="train")
val_dataloader = get_dataloader(DATA_PATH / "test", tag="validation")


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class FullModel(pl.LightningModule):
    def __post_init__(self):
        super().__init__()

        self.satellite_predictor = XResUNet(
            input_size=(64, 64),
            history_steps=NUM_HIST_SAT_IMAGES,
            forecast_steps=NUM_FUTURE_SAT_IMAGES,
            pretrained=False,
        )

        # Do this at the end of __post_init__ to capture model topology to wandb:
        self.save_hyperparameters()

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> dict[str, torch.Tensor]:
        historical_sat = x[BatchKey.hrvsatellite_actual][:, :NUM_HIST_SAT_IMAGES, 0]
        predicted_sat = self.satellite_predictor(historical_sat)
        return dict(predicted_sat=predicted_sat)

    def training_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        return self.validation_step(batch, batch_idx)

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        tag = "train" if self.training else "validation"
        network_out = self(batch)
        predicted_sat = network_out["predicted_sat"]
        actual_sat = batch[BatchKey.hrvsatellite_actual][:, NUM_HIST_SAT_IMAGES:, 0]
        sat_mse_loss = F.mse_loss(predicted_sat, actual_sat)
        self.log(f"{tag}/sat_mse", sat_mse_loss)

        # MS-SSIM. Requires images to be de-normalised:
        actual_sat_denorm = (actual_sat * SAT_STD["HRV"]) + SAT_MEAN["HRV"]
        predicted_sat_denorm = (predicted_sat * SAT_STD["HRV"]) + SAT_MEAN["HRV"]
        ms_ssim_loss = 1 - ms_ssim(
            predicted_sat_denorm,
            actual_sat_denorm,
            data_range=1023,
            size_average=True,  # Return a scalar.
            win_size=3,  # ClimateHack folks used win_size=3.
        )
        self.log(f"{tag}/ms_ssim", ms_ssim_loss)
        self.log(f"{tag}/ms_ssim+sat_mse", ms_ssim_loss + sat_mse_loss)

        return dict(
            loss=ms_ssim_loss + sat_mse_loss,
            predicted_sat=predicted_sat,
            actual_sat=actual_sat,
        )

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-4)
        return optimizer


model = FullModel()

wandb_logger = WandbLogger(
    name="021.03: MS-SSIM+SAT_MSE. LR=1e-4. ClimateHack satellite predictor. GCP-3",
    project="power_perceiver",
    entity="openclimatefix",
    log_model="all",
)

# log gradients, parameter histogram and model topology
# wandb_logger.watch(model, log="all")

trainer = pl.Trainer(
    gpus=[0],
    max_epochs=70,
    logger=wandb_logger,
    callbacks=[
        LogSatellitePlots(),
    ],
)

trainer.fit(
    model=model,
    train_dataloaders=train_dataloader,
    val_dataloaders=val_dataloader,
)



# General imports
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path

import einops

# ML imports
import matplotlib.pyplot as plt
import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from pytorch_lightning.loggers import WandbLogger
from torch import nn
from torch.utils import data

from power_perceiver.analysis.plot_timeseries import LogTimeseriesPlots

# power_perceiver imports
from power_perceiver.analysis.plot_tsne import LogTSNEPlot
from power_perceiver.consts import T0_IDX_30_MIN, BatchKey
from power_perceiver.load_prepared_batches.data_sources import GSP, PV, HRVSatellite, Sun
from power_perceiver.load_prepared_batches.prepared_dataset import PreparedDataset
from power_perceiver.np_batch_processor import EncodeSpaceTime, Topography
from power_perceiver.np_batch_processor.align_gsp_to_5_min import GSP5Min
from power_perceiver.pytorch_modules.query_generator import GSPQueryGenerator, PVQueryGenerator
from power_perceiver.pytorch_modules.satellite_processor import HRVSatelliteProcessor
from power_perceiver.pytorch_modules.self_attention import MultiLayerTransformerEncoder
from power_perceiver.transforms.pv import PVPowerRollingWindow
from power_perceiver.transforms.satellite import PatchSatellite
from power_perceiver.xr_batch_processor import (
    AlignGSPTo5Min,
    ReduceNumPVSystems,
    ReduceNumTimesteps,
    SelectPVSystemsNearCenterOfImage,
)

plt.rcParams["figure.figsize"] = (18, 10)
plt.rcParams["figure.facecolor"] = "white"

DATA_PATH = Path("/home/jack/data/v15")
#  "/mnt/storage_ssd_4tb/data/ocf/solar_pv_nowcasting/nowcasting_dataset_pipeline/"
#  "prepared_ML_training_data/v15/"
assert DATA_PATH.exists()

D_MODEL = 128
N_HEADS = 16
T0_IDX_5_MIN_TRAINING = 3
T0_IDX_5_MIN_VALIDATION = 12


def get_dataloader(data_path: Path, tag: str) -> data.DataLoader:
    assert tag in ["train", "validation"]
    assert data_path.exists()

    xr_batch_processors = [
        SelectPVSystemsNearCenterOfImage(),
        ReduceNumPVSystems(requested_num_pv_systems=8),
        AlignGSPTo5Min(),
    ]
    if tag == "train":
        xr_batch_processors.append(ReduceNumTimesteps(keys=(HRVSatellite, PV, GSP5Min, Sun)))

    dataset = PreparedDataset(
        data_path=data_path,
        data_loaders=[
            HRVSatellite(
                transforms=[
                    PatchSatellite(),
                ]
            ),
            PV(transforms=[PVPowerRollingWindow()]),
            Sun(),
            GSP(),
        ],
        xr_batch_processors=xr_batch_processors,
        np_batch_processors=[
            EncodeSpaceTime(),
            Topography("/home/jack/europe_dem_2km_osgb.tif"),
        ],
    )

    def seed_rngs(worker_id: int):
        """Set different random seed per worker."""
        worker_info = torch.utils.data.get_worker_info()
        for xr_batch_processor in worker_info.dataset.xr_batch_processors:
            if getattr(xr_batch_processor, "rng", None):
                xr_batch_processor.rng = np.random.default_rng(seed=42 + worker_id)

    dataloader = data.DataLoader(
        dataset,
        batch_size=None,
        num_workers=12,
        pin_memory=True,
        worker_init_fn=seed_rngs,
    )

    return dataloader


train_dataloader = get_dataloader(DATA_PATH / "train", tag="train")
val_dataloader = get_dataloader(DATA_PATH / "test", tag="validation")


def maybe_pad_with_zeros(tensor: torch.Tensor, requested_dim: int) -> torch.Tensor:
    num_zeros_to_pad = requested_dim - tensor.shape[-1]
    assert num_zeros_to_pad >= 0, f"{requested_dim=}, {tensor.shape=}"
    if num_zeros_to_pad > 0:
        zero_padding_shape = tensor.shape[:2] + (num_zeros_to_pad,)
        zero_padding = torch.zeros(*zero_padding_shape, dtype=tensor.dtype, device=tensor.device)
        tensor = torch.concat((tensor, zero_padding), dim=2)
    return tensor


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class SatelliteTransformer(nn.Module):
    """Infers a single timestep of PV power and GSP power at a time.

    Currently just uses HRV satellite imagery as the input. In the near future it could also
    use NWP temperature, wind speed & precipitation, and absolute geo position.
    """

    # Params for Perceiver
    # byte_array and query will be automatically padded with zeros to get to d_model.
    # Set d_model to be divisible by `num_heads`.
    d_model: int = D_MODEL
    pv_system_id_embedding_dim: int = 16
    num_heads: int = N_HEADS
    dropout: float = 0.0
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 4

    def __post_init__(self):
        super().__init__()
        self.hrvsatellite_processor = HRVSatelliteProcessor()

        NUM_PV_SYSTEMS = 1400
        NUM_GSPS = 360
        id_embedding = nn.Embedding(
            num_embeddings=NUM_PV_SYSTEMS + NUM_GSPS,
            embedding_dim=self.pv_system_id_embedding_dim,
        )

        self.pv_query_generator = PVQueryGenerator(
            pv_system_id_embedding=id_embedding,
            t0_idx_5_min_training=T0_IDX_5_MIN_TRAINING,
            t0_idx_5_min_validation=T0_IDX_5_MIN_VALIDATION,
        )

        self.gsp_query_generator = GSPQueryGenerator(gsp_id_embedding=id_embedding)

        self.transformer_encoder = MultiLayerTransformerEncoder(
            d_model=self.d_model,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> dict[str, torch.Tensor]:
        # Reshape so each timestep is considered a different example:
        num_examples, num_timesteps = x[BatchKey.pv].shape[:2]
        x = deepcopy(x)  # TODO: Fix this hack, which fixes the issue
        # that our reshaping affects the time_transformer, too!
        for batch_key in (
            BatchKey.gsp_5_min_time_utc_fourier,
            BatchKey.pv_time_utc_fourier,
            BatchKey.hrvsatellite_solar_azimuth,
            BatchKey.hrvsatellite_solar_elevation,
            BatchKey.hrvsatellite_actual,
            BatchKey.hrvsatellite_time_utc_fourier,
        ):
            x[batch_key] = einops.rearrange(x[batch_key], "example time ... -> (example time) ...")

        # Process satellite data and queries:
        pv_query = self.pv_query_generator(x, for_satellite_transformer=True)
        gsp_query = self.gsp_query_generator(x, for_satellite_transformer=True)
        satellite_data = self.hrvsatellite_processor(x)

        # Pad with zeros if necessary to get up to self.d_model:
        pv_query = maybe_pad_with_zeros(pv_query, requested_dim=self.d_model)
        gsp_query = maybe_pad_with_zeros(gsp_query, requested_dim=self.d_model)
        satellite_data = maybe_pad_with_zeros(satellite_data, requested_dim=self.d_model)

        # Mask the NaN GSP queries. True or non-zero value indicates value will be ignored.
        mask = torch.concat(
            (
                torch.zeros_like(pv_query[:, :, 0]),
                torch.isnan(gsp_query[:, :, -1]),  # The time fourier features are last in dim 2.
                torch.zeros_like(satellite_data[:, :, 0]),
            ),
            dim=1,
        )

        # Now set NaN GSP queries to zero. They'll be ignored :)
        gsp_query = torch.nan_to_num(gsp_query, nan=0)

        # Prepare the attention input and run through the transformer_encoder:
        attn_input = torch.concat((pv_query, gsp_query, satellite_data), dim=1)
        attn_output = attn_input + self.transformer_encoder(attn_input, src_key_padding_mask=mask)

        # Reshape to (example time element d_model):
        attn_output = einops.rearrange(
            attn_output,
            "(example time) ... -> example time ...",
            example=num_examples,
            time=num_timesteps,
        )

        # Select the elements of the output which correspond to the query:
        gsp_start_idx = pv_query.shape[1]
        gsp_end_idx = gsp_start_idx + gsp_query.shape[1]
        pv_attn_out = attn_output[:, :, :gsp_start_idx]
        gsp_attn_out = attn_output[:, :, gsp_start_idx:gsp_end_idx]

        return {
            "pv_attn_out": pv_attn_out,  # shape: (example, n_pv_systems, d_model)
            "gsp_attn_out": gsp_attn_out,  # shape: (example, 1, d_model)
        }


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class FullModel(pl.LightningModule):
    d_model: int = D_MODEL + N_HEADS  # + N_HEADS for historical PV
    pv_system_id_embedding_dim: int = 16
    num_heads: int = N_HEADS
    dropout: float = 0.1
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 4

    def __post_init__(self):
        super().__init__()
        self.satellite_transformer = SatelliteTransformer()

        self.time_transformer = MultiLayerTransformerEncoder(
            d_model=self.d_model,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

        self.pv_output_module = nn.Sequential(
            nn.Linear(in_features=self.d_model, out_features=self.d_model),
            nn.GELU(),
            nn.Linear(in_features=self.d_model, out_features=1),
            nn.ReLU(),  # ReLU output guarantees that we can't predict negative PV power!
        )

        # Do this at the end of __post_init__ to capture model topology to wandb:
        self.save_hyperparameters()

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> dict[str, torch.Tensor]:
        sat_trans_out = self.satellite_transformer(x)
        pv_attn_out = sat_trans_out["pv_attn_out"]  # Shape: (example time n_pv_systems d_model)
        gsp_attn_out = sat_trans_out["gsp_attn_out"]

        # Concatenate historical PV
        t0_idx_5_min = T0_IDX_5_MIN_TRAINING if self.training else T0_IDX_5_MIN_VALIDATION
        historical_pv = torch.zeros_like(x[BatchKey.pv])  # Shape: (example, time, n_pv_systems)
        historical_pv[:, : t0_idx_5_min + 1] = x[BatchKey.pv][:, : t0_idx_5_min + 1]
        historical_pv = historical_pv.unsqueeze(-1)  # Shape: (example, time, n_pv_systems, 1)
        # Now append a "marker" to indicate which timesteps are history:
        hist_pv_marker = torch.zeros_like(historical_pv)
        hist_pv_marker[:, : t0_idx_5_min + 1] = 1
        pv_attn_out = torch.concat((pv_attn_out, historical_pv, hist_pv_marker), dim=3)

        # Reshape pv and gsp attention outputs so each timestep an each pv system is
        # seen as a separate element into the `time transformer`.
        n_timesteps, n_pv_systems = pv_attn_out.shape[1:3]
        REARRANGE_STR = "example time n_pv_systems d_model -> example (time n_pv_systems) d_model"
        pv_attn_out = einops.rearrange(pv_attn_out, REARRANGE_STR)
        pv_attn_out = maybe_pad_with_zeros(pv_attn_out, requested_dim=self.d_model)
        gsp_attn_out = einops.rearrange(gsp_attn_out, REARRANGE_STR, n_pv_systems=1)
        gsp_attn_out = maybe_pad_with_zeros(gsp_attn_out, requested_dim=self.d_model)
        n_pv_elements = pv_attn_out.shape[1]

        # Get historical PV
        # pv_query_generator = self.satellite_transformer.pv_query_generator
        # historical_pv = pv_query_generator(x, for_satellite_transformer=False)
        # historical_pv = maybe_pad_with_zeros(historical_pv, requested_dim=self.d_model)

        # Get GSP query
        gsp_query_generator = self.satellite_transformer.gsp_query_generator
        gsp_query = gsp_query_generator(x, for_satellite_transformer=False)
        gsp_query = maybe_pad_with_zeros(gsp_query, requested_dim=self.d_model)

        # Concatenate all the things we're going to feed into the "time transformer":
        time_attn_in = (
            pv_attn_out,
            gsp_attn_out,
            # historical_pv,
            gsp_query,
        )
        time_attn_in = torch.concat(time_attn_in, dim=1)
        time_attn_out = self.time_transformer(time_attn_in)

        power_out = self.pv_output_module(time_attn_out)  # (example, total_num_elements, 1)

        # Reshape the PV power predictions
        predicted_pv_power = power_out[:, :n_pv_elements]
        predicted_pv_power = einops.rearrange(
            predicted_pv_power,
            "example (time n_pv_systems) 1 -> example time n_pv_systems",
            time=n_timesteps,
            n_pv_systems=n_pv_systems,
        )

        # GSP power. There's just 1 GSP.
        n_gsp_elements = gsp_query.shape[1]
        predicted_gsp_power = power_out[:, -n_gsp_elements:, 0]

        return dict(
            predicted_pv_power=predicted_pv_power,  # Shape: (example time n_pv_systems)
            predicted_gsp_power=predicted_gsp_power,  # Shape: (example, time)
        )

    def training_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        return self.validation_step(batch, batch_idx)

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        tag = "train" if self.training else "validation"
        network_out = self(batch)
        predicted_pv_power = network_out["predicted_pv_power"]
        predicted_gsp_power = network_out["predicted_gsp_power"]
        actual_pv_power = batch[BatchKey.pv]
        actual_pv_power = torch.where(
            batch[BatchKey.pv_mask].unsqueeze(1),
            actual_pv_power,
            torch.tensor(0.0, dtype=actual_pv_power.dtype, device=actual_pv_power.device),
        )

        # PV power loss:
        pv_mse_loss = F.mse_loss(predicted_pv_power, actual_pv_power)
        t0_idx_5_min = T0_IDX_5_MIN_TRAINING if self.training else T0_IDX_5_MIN_VALIDATION

        pv_nmae_loss = F.l1_loss(
            predicted_pv_power[:, t0_idx_5_min + 1 :], actual_pv_power[:, t0_idx_5_min + 1 :]
        )
        self.log(f"{tag}/pv_mse", pv_mse_loss)
        self.log(f"{tag}/pv_nmae", pv_nmae_loss)

        # GSP power loss:
        actual_gsp_power = batch[BatchKey.gsp]
        gsp_mse_loss = F.mse_loss(predicted_gsp_power, actual_gsp_power)
        gsp_nmae_loss = F.l1_loss(
            predicted_gsp_power[:, T0_IDX_30_MIN + 1 :], actual_gsp_power[:, T0_IDX_30_MIN + 1 :]
        )
        self.log(f"{tag}/gsp_mse", gsp_mse_loss)
        self.log(f"{tag}/gsp_nmae", gsp_nmae_loss)

        # Total MSE loss:
        total_mse_loss = pv_mse_loss + gsp_mse_loss
        self.log(f"{tag}/total_mse", total_mse_loss)

        # Total NMAE loss:
        total_nmae_loss = pv_nmae_loss + gsp_nmae_loss
        self.log(f"{tag}/total_nmae", total_nmae_loss)

        return {
            "loss": total_mse_loss,
            "pv_mse_loss": pv_mse_loss,
            "gsp_mse_loss": gsp_mse_loss,
            "pv_nmae_loss": pv_nmae_loss,
            "gsp_nmae_loss": gsp_nmae_loss,
            "predicted_gsp_power": predicted_gsp_power,
            "actual_gsp_power": actual_gsp_power,
            "gsp_time_utc": batch[BatchKey.gsp_time_utc],
            "actual_pv_power": actual_pv_power,
            "predicted_pv_power": predicted_pv_power,
        }

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=5e-5)
        return optimizer


model = FullModel()

wandb_logger = WandbLogger(
    name=(
        "020.11: GCP! Mark hist. MSE objective. D_MODEL=128. 4 TT layers."
        " Concat hist PV. 12 timesteps during training. LR=5e-5"
    ),
    project="power_perceiver",
    entity="openclimatefix",
    log_model="all",
)

# log gradients, parameter histogram and model topology
# wandb_logger.watch(model, log="all")

trainer = pl.Trainer(
    gpus=[0],
    max_epochs=70,
    logger=wandb_logger,
    callbacks=[
        LogTimeseriesPlots(),
        LogTSNEPlot(query_generator_name="satellite_transformer.pv_query_generator"),
    ],
)

trainer.fit(
    model=model,
    train_dataloaders=train_dataloader,
    val_dataloaders=val_dataloader,
)



"""At the moment, it's actually quite tricky to align the topo data with the satellite data
(because xarray.align doesn't like the fact that the satellite data's OSGB coords are each 2D.
So we have to reproject the topo data first, and then align it).

This will no longer be necessary if / when this is implemented:
https://github.com/openclimatefix/nowcasting_dataset/issues/642

The strategy is:

1. Load all the topo data into memory.
2. Reproject it from OSGB to geostationary CRS. We have to do this because the satellite
   data is natively in geostationary CRS. Yes, the satellite data comes with OSGB coords,
   but those are 2D, and xr.combine_by_coords doesn't work with 2D coords. So we have to
   reproject the topo data to geostationary so we end up with both the topo data and the
   satellite data having "normal" 1D coords.
3. Then, for each satellite example, use xarray.combine_by_coords twice:
   1. An "outer join" to get a DataArray with the coords from both the satellite
      data and the topo data. This results in topo data with lots of NaNs! So we
      forward-fill and backwards-fill the NaNs.
    2. A "right join" to select just the coords that exactly match the satellite data's
       geostationary coords.
"""

from dataclasses import dataclass

import cartopy.crs as ccrs
import numpy as np
import pyproj
import pyresample
import rioxarray
import xarray as xr

from power_perceiver.consts import BatchKey
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import NumpyBatch


@dataclass
class Topography:
    topo_filename: str

    def __post_init__(self):
        self.topo = load_topo_data(self.topo_filename)
        self.topo = reproject_topo_data_from_osgb_to_geostationary(self.topo)
        self.topo = self.topo.fillna(0)
        self.topo_mean = self.topo.mean().item()
        self.topo_std = self.topo.std().item()

    def __call__(self, np_batch: NumpyBatch) -> NumpyBatch:
        if BatchKey.hrvsatellite_x_geostationary in np_batch:
            # Recreate an xr.DataArray of the satellite data. This is required so we can
            # use xr.combine_by_coords to align the topo data with the satellite data.
            hrvsatellite_data_array = xr.DataArray(
                # We're not actually interested in the image. But xarray won't make an
                # empty DataArray without data in the right shape.
                # There's nothing special about the x_osgb data. It's just convenient because
                # it's the right shape!
                np_batch[BatchKey.hrvsatellite_x_osgb],
                dims=("example", "y", "x"),
                coords={
                    "y_geostationary": (
                        ("example", "y"),
                        np_batch[BatchKey.hrvsatellite_y_geostationary],
                    ),
                    "x_geostationary": (
                        ("example", "x"),
                        np_batch[BatchKey.hrvsatellite_x_geostationary],
                    ),
                },
            )
            hrvsatellite_surface_height = _get_surface_height_for_satellite(
                surface_height=self.topo, satellite=hrvsatellite_data_array
            )
            hrvsatellite_surface_height = self._normalise(hrvsatellite_surface_height)
            np_batch[BatchKey.hrvsatellite_surface_height] = hrvsatellite_surface_height

        return np_batch

    def _normalise(self, data: np.ndarray) -> np.ndarray:
        data = np.nan_to_num(data, nan=0.0)
        data -= self.topo_mean
        data /= self.topo_std
        return data


def load_topo_data(filename: str) -> xr.DataArray:
    topo = rioxarray.open_rasterio(filename=filename, parse_coordinates=True, masked=True)

    # `band` and `spatial_ref` don't appear to hold any useful info. So get rid of them:
    topo = topo.isel(band=0)
    topo = topo.drop_vars(["spatial_ref", "band"])

    # Use our standard naming:
    topo = topo.rename({"x": "x_osgb", "y": "y_osgb"})

    # Select Western Europe:
    topo = topo.sel(x_osgb=slice(-300_000, 1_500_000), y_osgb=slice(1_300_000, -800_000))

    return topo


def reproject_topo_data_from_osgb_to_geostationary(topo: xr.DataArray) -> xr.DataArray:
    topo_osgb_area_def = _get_topo_osgb_area_def(topo)
    topo_geostationary_area_def = _get_topo_geostationary_area_def(topo)
    topo_image = pyresample.image.ImageContainerQuick(topo.values, topo_osgb_area_def)
    topo_image_resampled = topo_image.resample(topo_geostationary_area_def)
    return _get_data_array_of_resampled_topo_image(topo_image_resampled)


def _get_topo_osgb_area_def(topo: xr.DataArray) -> pyresample.geometry.AreaDefinition:
    # Get AreaDefinition of the OSGB topographical data:
    osgb = ccrs.OSGB(approx=False)
    return pyresample.create_area_def(
        area_id="OSGB",
        projection=osgb.proj4_params,
        shape=topo.shape,  # y, x
        area_extent=(
            topo.x_osgb[0].item(),  # lower_left_x
            topo.y_osgb[-1].item(),  # lower_left_y
            topo.x_osgb[-1].item(),  # upper_right_x
            topo.y_osgb[0].item(),  # upper_right_y
        ),
    )


def _get_topo_geostationary_area_def(topo: xr.DataArray) -> pyresample.geometry.AreaDefinition:
    # Get the geostationary boundaries of the topo data:
    OSGB_EPSG_CODE = 27700
    GEOSTATIONARY_PROJ = {
        "proj": "geos",
        "lon_0": 9.5,
        "h": 35785831,
        "x_0": 0,
        "y_0": 0,
        "a": 6378169,
        "rf": 295.488065897014,
        "no_defs": None,
        "type": "crs",
    }
    osgb_to_geostationary = pyproj.Transformer.from_crs(
        crs_from=OSGB_EPSG_CODE, crs_to=GEOSTATIONARY_PROJ
    ).transform
    lower_left_geos = osgb_to_geostationary(xx=topo.x_osgb[0], yy=topo.y_osgb[-1])
    upper_right_geos = osgb_to_geostationary(xx=topo.x_osgb[-1], yy=topo.y_osgb[0])
    shape = (topo.shape[0] * 2, topo.shape[1] * 2)  # Oversample to make sure we don't loose info.
    return pyresample.create_area_def(
        area_id="msg_seviri_rss_1km",
        projection=GEOSTATIONARY_PROJ,
        shape=shape,
        # lower_left_x, lower_left_y, upper_right_x, upper_right_y:
        area_extent=lower_left_geos + upper_right_geos,
    )


def _get_data_array_of_resampled_topo_image(
    topo_image_resampled: pyresample.image.ImageContainer,
) -> xr.DataArray:
    (
        lower_left_x,
        lower_left_y,
        upper_right_x,
        upper_right_y,
    ) = topo_image_resampled.geo_def.area_extent
    return xr.DataArray(
        topo_image_resampled.image_data,
        coords=(
            (
                "y",
                np.linspace(
                    start=upper_right_y, stop=lower_left_y, num=topo_image_resampled.shape[0]
                ),
            ),
            (
                "x",
                np.linspace(
                    start=lower_left_x, stop=upper_right_x, num=topo_image_resampled.shape[1]
                ),
            ),
        ),
    )


def _get_surface_height_for_satellite(
    surface_height: xr.DataArray, satellite: xr.DataArray
) -> np.ndarray:
    num_examples = satellite.shape[0]
    surface_height = surface_height.rename("surface_height")
    surface_height_for_batch = np.full_like(satellite.values, fill_value=np.NaN)
    for example_idx in range(num_examples):
        satellite_example = satellite.isel(example=example_idx)
        msg = "Satellite imagery must start in the top-left!"
        assert satellite_example.y_geostationary[0] > satellite_example.y_geostationary[-1], msg
        assert satellite_example.x_geostationary[0] < satellite_example.x_geostationary[-1], msg

        satellite_example = satellite_example.rename(
            {"y_geostationary": "y", "x_geostationary": "x"}
        ).rename("sat")
        surface_height_for_example = surface_height.sel(
            y=slice(
                satellite_example.y[0],
                satellite_example.y[-1],
            ),
            x=slice(
                satellite_example.x[0],
                satellite_example.x[-1],
            ),
        )
        # Align by coordinates. This will result in lots of NaNs in the surface height data:
        aligned = xr.combine_by_coords(
            (surface_height_for_example, satellite_example), join="outer"
        )
        # Fill in the NaNs:
        surface_height_for_example = (
            aligned["surface_height"].ffill("x").ffill("y").bfill("x").bfill("y")
        )

        # Now select exactly the same coordinates from the surface height data as the satellite data
        aligned = xr.combine_by_coords((surface_height_for_example, satellite_example), join="left")

        surface_height_for_batch[example_idx] = aligned["surface_height"].values

    # If we slightly ran off the edge of the topo data then we'll get NaNs.
    # TODO: Enlarge topo data so we never get NaNs!
    surface_height_for_batch = np.nan_to_num(surface_height_for_batch, nan=0)

    return surface_height_for_batch



import datetime
import logging

from power_perceiver.consts import Location
from power_perceiver.load_raw.raw_dataset import RawDataset

_log = logging.getLogger(__name__)


class NationalPVDataset(RawDataset):
    """Only pass in a single `data_source_combo` (with GSP as the first entry.)"""

    def _epoch_start(self) -> None:
        super()._epoch_start()
        self._reset()
        self._gsp_data_source = self.data_source_combos[self._gsp_combo_name][0]
        self._num_gsps_total = self._gsp_data_source.num_gsps

    def _choose_osgb_location(self, chosen_combo_name: str) -> Location:
        loc = self._gsp_data_source.get_osgb_location_for_gsp_idx(self._gsp_idx)
        if self._gsp_idx < self._num_gsps_total - 1:
            self._gsp_idx += 1
        else:
            self._reset()
        return loc

    def _reset(self) -> None:
        self._gsp_idx = 0
        self._t0_datetime = super()._choose_t0_datetime(self._gsp_combo_name)
        _log.info("Resetting NationalPVDataset!")

    def _choose_t0_datetime(self, chosen_combo_name: str) -> datetime.datetime:
        return self._t0_datetime

    @property
    def _gsp_combo_name(self) -> str:
        return list(self.data_source_combos.keys())[0]



# General imports
from dataclasses import dataclass
from pathlib import Path

# ML imports
import einops
import matplotlib.pyplot as plt
import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from pytorch_lightning.loggers import WandbLogger
from torch import nn
from torch.utils import data

# power_perceiver imports
from power_perceiver.analysis.plot_timeseries import LogTimeseriesPlots
from power_perceiver.analysis.plot_tsne import LogTSNEPlot
from power_perceiver.consts import BatchKey
from power_perceiver.load_prepared_batches.data_sources import PV, HRVSatellite, Sun
from power_perceiver.load_prepared_batches.prepared_dataset import PreparedDataset
from power_perceiver.np_batch_processor import EncodeSpaceTime, Topography
from power_perceiver.pytorch_modules.query_generator import QueryGenerator
from power_perceiver.pytorch_modules.satellite_processor import HRVSatelliteProcessor
from power_perceiver.pytorch_modules.self_attention import MultiLayerTransformerEncoder
from power_perceiver.transforms.pv import PVPowerRollingWindow
from power_perceiver.transforms.satellite import PatchSatellite
from power_perceiver.xr_batch_processor import (
    ReduceNumPVSystems,
    ReduceNumTimesteps,
    SelectPVSystemsNearCenterOfImage,
)

plt.rcParams["figure.figsize"] = (18, 10)
plt.rcParams["figure.facecolor"] = "white"

DATA_PATH = Path(
    "/mnt/storage_ssd_4tb/data/ocf/solar_pv_nowcasting/nowcasting_dataset_pipeline/"
    "prepared_ML_training_data/v15/"
)
assert DATA_PATH.exists()


def get_dataloader(data_path: Path, tag: str) -> data.DataLoader:
    assert tag in ["train", "validation"]
    assert data_path.exists()

    xr_batch_processors = [
        SelectPVSystemsNearCenterOfImage(),
        ReduceNumPVSystems(requested_num_pv_systems=8),
    ]

    if tag == "train":
        xr_batch_processors.append(ReduceNumTimesteps(requested_timesteps=4))

    dataset = PreparedDataset(
        data_path=data_path,
        data_loaders=[
            HRVSatellite(
                transforms=[
                    PatchSatellite(),
                ]
            ),
            PV(transforms=[PVPowerRollingWindow()]),
            Sun(),
        ],
        xr_batch_processors=xr_batch_processors,
        np_batch_processors=[
            EncodeSpaceTime(),
            Topography("/home/jack/europe_dem_2km_osgb.tif"),
        ],
    )

    def seed_rngs(worker_id: int):
        """Set different random seed per worker."""
        worker_info = torch.utils.data.get_worker_info()
        for xr_batch_processor in worker_info.dataset.xr_batch_processors:
            if getattr(xr_batch_processor, "rng", None):
                xr_batch_processor.rng = np.random.default_rng(seed=42 + worker_id)

    dataloader = data.DataLoader(
        dataset,
        batch_size=None,
        num_workers=16,
        pin_memory=True,
        worker_init_fn=seed_rngs,
    )

    return dataloader


train_dataloader = get_dataloader(DATA_PATH / "train", tag="train")
val_dataloader = get_dataloader(DATA_PATH / "test", tag="validation")


def maybe_pad_with_zeros(tensor: torch.Tensor, requested_dim: int) -> torch.Tensor:
    num_zeros_to_pad = requested_dim - tensor.shape[-1]
    assert num_zeros_to_pad >= 0, f"{requested_dim=}, {tensor.shape=}"
    if num_zeros_to_pad > 0:
        zero_padding_shape = tensor.shape[:2] + (num_zeros_to_pad,)
        zero_padding = torch.zeros(*zero_padding_shape, dtype=tensor.dtype, device=tensor.device)
        tensor = torch.concat((tensor, zero_padding), dim=2)
    return tensor


@dataclass(
    eq=False
)  # See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
class Model(pl.LightningModule):
    # Params for Perceiver
    query_dim: int = (
        36  # byte_array and query will be automatically padded with zeros to get to this size.
    )
    num_fourier_features: int = 16  # TOTAL for both x and y
    pv_system_id_embedding_dim: int = 16
    num_heads: int = 6
    dropout: float = 0.0
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 12

    # Other params:
    # Probably keep this at zero while using MultiLayerTransformerEncoder
    num_elements_query_padding: int = 0

    def __post_init__(self):
        super().__init__()
        self.hrvsatellite_processor = HRVSatelliteProcessor()

        self.query_generator = QueryGenerator(
            num_fourier_features=self.num_fourier_features,  # TOTAL (for both x and y)
            pv_system_id_embedding_dim=self.pv_system_id_embedding_dim,
            num_elements_query_padding=self.num_elements_query_padding,
        )

        self.transformer_encoder = MultiLayerTransformerEncoder(
            d_model=self.query_dim,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

        self.output_module = nn.Sequential(
            nn.Linear(in_features=self.query_dim, out_features=self.query_dim),
            nn.GELU(),
            nn.Linear(in_features=self.query_dim, out_features=1),
        )

        # Do this at the end of __post_init__ to capture model topology:
        self.save_hyperparameters()

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
        original_batch_size = x[BatchKey.pv].shape[0]
        byte_array = self.hrvsatellite_processor(x)
        query = self.query_generator(x)

        # Pad with zeros if necessary to get up to self.query_dim:
        byte_array = maybe_pad_with_zeros(byte_array, requested_dim=self.query_dim)
        query = maybe_pad_with_zeros(query, requested_dim=self.query_dim)

        # Prepare the attention input and run through the transformer_encoder:
        attn_input = torch.concat((byte_array, query), dim=1)
        attn_output = self.transformer_encoder(attn_input)

        # Select the elements of the output which correspond to the query:
        out = attn_output[:, byte_array.shape[1] :]

        out = self.output_module(out)

        # Reshape back to (batch_size, n_timesteps, ...)
        return einops.rearrange(
            out,
            "(batch_size n_timesteps) ... -> batch_size n_timesteps ...",
            batch_size=original_batch_size,
        )

    def _training_or_validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int, tag: str
    ) -> dict[str, object]:
        """
        Args:
            batch: The training or validation batch.  A dictionary.
            tag: Either "train" or "validation"
            batch_idx: The index of the batch.
        """
        actual_pv_power = batch[BatchKey.pv]
        actual_pv_power = torch.where(
            batch[BatchKey.pv_mask].unsqueeze(1),
            actual_pv_power,
            torch.tensor(0.0, dtype=actual_pv_power.dtype, device=actual_pv_power.device),
        )

        predicted_pv_power = self(batch).squeeze()
        mse_loss = F.mse_loss(predicted_pv_power, actual_pv_power)

        self.log(f"{tag}/mse", mse_loss)

        return {
            "loss": mse_loss,
            "predicted_pv_power": predicted_pv_power,
        }

    def training_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        return self._training_or_validation_step(batch=batch, batch_idx=batch_idx, tag="train")

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        return self._training_or_validation_step(batch=batch, batch_idx=batch_idx, tag="validation")

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=5e-5)
        return optimizer


model = Model()

wandb_logger = WandbLogger(
    project="power_perceiver",
    entity="openclimatefix",
    log_model="all",
)

# log gradients, parameter histogram and model topology
wandb_logger.watch(model, log="all")

trainer = pl.Trainer(
    gpus=[4],
    max_epochs=70,
    logger=wandb_logger,
    callbacks=[
        LogTimeseriesPlots(),
        LogTSNEPlot(query_generator_name="query_generator"),
    ],
)

trainer.fit(
    model=model,
    train_dataloaders=train_dataloader,
    val_dataloaders=val_dataloader,
)



import numpy as np
import pytest

from power_perceiver.production.batch import GSP, NWP, PV, HRVSatellite


@pytest.fixture()
def pv():

    batch_size = 4
    timesteps = 3
    pv_systems = 2

    pv = np.random.random([batch_size, timesteps, pv_systems])
    row_number = np.random.random([batch_size, pv_systems])
    pv_y_osgb = np.random.random([batch_size, pv_systems])
    pv_x_osgb = np.random.random([batch_size, pv_systems])
    pv_time_utc = np.random.random([batch_size, timesteps])

    return PV(
        pv=pv,
        pv_t0_idx=10,
        pv_system_row_number=row_number,
        pv_y_osgb=pv_y_osgb,
        pv_x_osgb=pv_x_osgb,
        pv_time_utc=pv_time_utc,
    )


@pytest.fixture()
def gsp():

    batch_size = 4
    timesteps = 3

    gsp = np.random.random([batch_size, timesteps, 1])
    gsp_id = np.random.random([batch_size])
    gsp_y_osgb = np.random.random([batch_size])
    gsp_x_osgb = np.random.random([batch_size])
    gsp_time_utc = np.random.random([batch_size, timesteps])

    return GSP(
        gsp=gsp,
        gsp_t0_idx=10,
        gsp_id=gsp_id,
        gsp_y_osgb=gsp_y_osgb,
        gsp_x_osgb=gsp_x_osgb,
        gsp_time_utc=gsp_time_utc,
    )


@pytest.fixture()
def nwp():

    batch_size = 4
    timesteps = 3
    channels = 2
    x_steps = 5
    y_steps = 6

    nwp = np.random.random([batch_size, timesteps, channels, y_steps, x_steps])
    nwp_target_time_utc = np.random.random([batch_size, timesteps])
    nwp_y_osgb = np.random.random([batch_size, y_steps])
    nwp_x_osgb = np.random.random([batch_size, x_steps])
    nwp_init_time_utc = np.random.random([batch_size, timesteps])

    return NWP(
        nwp=nwp,
        nwp_t0_idx=10,
        nwp_target_time_utc=nwp_target_time_utc,
        nwp_y_osgb=nwp_y_osgb,
        nwp_x_osgb=nwp_x_osgb,
        nwp_init_time_utc=nwp_init_time_utc,
    )


@pytest.fixture()
def hrv_satellite():

    batch_size = 4
    timesteps = 3
    channels = 2
    x_steps = 5
    y_steps = 6

    hrvsatellite_actual = np.random.random([batch_size, timesteps, channels, y_steps, x_steps])
    hrvsatellite_x_geostationary = np.random.random([batch_size, y_steps, x_steps])
    hrvsatellite_y_geostationary = np.random.random([batch_size, y_steps, x_steps])
    hrvsatellite_y_osgb = np.random.random([batch_size, y_steps, x_steps])
    hrvsatellite_x_osgb = np.random.random([batch_size, y_steps, x_steps])
    hrvsatellite_time_utc = np.random.random([batch_size, timesteps])

    return HRVSatellite(
        hrvsatellite_actual=hrvsatellite_actual,
        hrvsatellite_x_geostationary=hrvsatellite_y_geostationary,
        hrvsatellite_y_geostationary=hrvsatellite_x_geostationary,
        hrvsatellite_y_osgb=hrvsatellite_y_osgb,
        hrvsatellite_x_osgb=hrvsatellite_x_osgb,
        hrvsatellite_time_utc=hrvsatellite_time_utc,
        hrvsatellite_t0_idx=10,
    )



"""Satellite transforms."""

from dataclasses import dataclass

import xarray as xr

from power_perceiver.load_prepared_batches.data_sources.satellite import _set_sat_coords


@dataclass
class PatchSatellite:
    """Appends a `patch` dimension to the satellite data.

    The `patch` dim will be x_patch_size_pixels * y_patch_size_pixels in size.

    TODO: Given that we have to `patch` the satellite imagery using `einops`
    after `SatellitePredictor`, maybe we should get rid of this `PatchSatellite`
    class and always use `einops` to patch the satellite imagery.
    """

    y_patch_size_pixels: int = 4
    x_patch_size_pixels: int = 4

    def __call__(self, dataset: xr.Dataset) -> xr.Dataset:
        dataset = dataset.coarsen(y=self.y_patch_size_pixels, x=self.x_patch_size_pixels)
        dataset = dataset.construct(y=("y", "y_patch"), x=("x", "x_patch"))
        dataset = dataset.stack(patch=("y_patch", "x_patch"))

        # Downsample spatial coordinates
        # TODO: I think this can be replaced by passing `coord_func='mean'` into `dataset.coarsen`
        COORD_NAMES = ["x_geostationary", "x_osgb", "y_geostationary", "y_osgb"]
        dataset[COORD_NAMES] = dataset[COORD_NAMES].mean(dim="patch")

        dataset = _set_sat_coords(dataset)

        return dataset



#!/usr/bin/env python3

import hydra
from omegaconf import DictConfig, OmegaConf


@hydra.main(config_path="../configs/", config_name="config")
def main(cfg: DictConfig) -> None:
    print(OmegaConf.to_yaml(cfg))


if __name__ == "__main__":
    main()



"""Functions operating on dates and times."""

import datetime
import itertools
from numbers import Number
from typing import Union

import numpy as np
import pandas as pd
import pvlib
import xarray as xr

from power_perceiver.geospatial import osgb_to_lat_lon


def get_dates(time_index: pd.DatetimeIndex) -> np.ndarray:  # noqa: D103
    return np.sort(np.unique(time_index.date))


def num_days(time_index: pd.DatetimeIndex) -> int:  # noqa: D103
    dates = get_dates(time_index)
    return len(dates)


def date_summary_str(time_index: pd.DatetimeIndex) -> str:
    """Convert to pd.DatetimeIndex to get prettier date string formatting."""
    time_index = pd.DatetimeIndex(time_index)
    return (
        f"there are {num_days(time_index):,d} days of data"
        f" from {time_index[0]} to {time_index[-1]}."
        f" A total of {len(time_index):,d} timesteps."
    )


def select_data_in_daylight(
    xr_data: Union[xr.Dataset, xr.DataArray], solar_elevation_threshold_degrees: Number = 10
) -> Union[xr.Dataset, xr.DataArray]:
    """Only select data where, for at least one of the four corners of the imagery,
    the Sun is at least `solar_elevation_threshold_degrees` above the horizon."""
    y_osgb = xr_data.y_osgb
    x_osgb = xr_data.x_osgb

    corners_osgb = [
        (
            x_osgb.isel(x_geostationary=x, y_geostationary=y).values,
            y_osgb.isel(x_geostationary=x, y_geostationary=y).values,
        )
        for x, y in itertools.product((0, -1), (0, -1))
    ]

    corners_osgb = pd.DataFrame(corners_osgb, columns=["x", "y"])

    lats, lons = osgb_to_lat_lon(x=corners_osgb.x, y=corners_osgb.y)

    elevation_for_all_corners = []
    for lat, lon in zip(lats, lons):
        solpos = pvlib.solarposition.get_solarposition(
            time=xr_data.time_utc,
            latitude=lat,
            longitude=lon,
        )
        elevation = solpos["elevation"]
        elevation_for_all_corners.append(elevation)

    elevation_for_all_corners = pd.concat(elevation_for_all_corners, axis="columns")
    max_elevation = elevation_for_all_corners.max(axis="columns")
    daylight_hours_mask = max_elevation >= solar_elevation_threshold_degrees
    return xr_data.isel(time_utc=daylight_hours_mask)


# From nowcasting_dataset.time:
def get_contiguous_time_periods(
    datetimes: pd.DatetimeIndex,
    min_seq_length: int,
    max_gap_duration: datetime.timedelta,
) -> pd.DataFrame:
    """Return a pd.DataFrame where each row records the boundary of a contiguous time period.

    Args:
      datetimes: pd.DatetimeIndex. Must be sorted.
      min_seq_length: Sequences of min_seq_length or shorter will be discarded.  Typically, this
        would be set to the `total_seq_length` of each machine learning example.
      max_gap_duration: If any pair of consecutive `datetimes` is more than `max_gap_duration`
        apart, then this pair of `datetimes` will be considered a "gap" between two contiguous
        sequences. Typically, `max_gap_duration` would be set to the sample period of
        the timeseries.

    Returns:
      pd.DataFrame where each row represents a single time period.  The pd.DataFrame
          has two columns: `start_dt` and `end_dt` (where 'dt' is short for 'datetime').
    """
    # Sanity checks.
    assert len(datetimes) > 0
    assert min_seq_length > 1
    assert datetimes.is_monotonic_increasing
    assert datetimes.is_unique

    # Find indices of gaps larger than max_gap:
    gap_mask = pd.TimedeltaIndex(np.diff(datetimes)) > max_gap_duration
    gap_indices = np.argwhere(gap_mask)[:, 0]

    # gap_indicies are the indices into dt_index for the timestep immediately before the gap.
    # e.g. if the datetimes at 12:00, 12:05, 18:00, 18:05 then gap_indicies will be [1].
    # So we add 1 to gap_indices to get segment_boundaries, an index into dt_index
    # which identifies the _start_ of each segment.
    segment_boundaries = gap_indices + 1

    # Capture the last segment of dt_index.
    segment_boundaries = np.concatenate((segment_boundaries, [len(datetimes)]))

    periods: list[dict[str, pd.Timestamp]] = []
    start_i = 0
    for next_start_i in segment_boundaries:
        n_timesteps = next_start_i - start_i
        if n_timesteps > min_seq_length:
            end_i = next_start_i - 1
            period = {"start_dt": datetimes[start_i], "end_dt": datetimes[end_i]}
            periods.append(period)
        start_i = next_start_i

    assert len(periods) > 0

    return pd.DataFrame(periods)


# From nowcasting_dataset.time:
def intersection_of_multiple_dataframes_of_periods(
    time_periods: list[pd.DataFrame],
) -> pd.DataFrame:
    """Find the intersection of a list of time periods.

    See the docstring of intersection_of_2_dataframes_of_periods() for more details.
    """
    assert len(time_periods) > 0
    if len(time_periods) == 1:
        return time_periods[0]
    intersection = intersection_of_2_dataframes_of_periods(time_periods[0], time_periods[1])
    for time_period in time_periods[2:]:
        intersection = intersection_of_2_dataframes_of_periods(intersection, time_period)
    return intersection


# From nowcasting_dataset.time:
def intersection_of_2_dataframes_of_periods(a: pd.DataFrame, b: pd.DataFrame) -> pd.DataFrame:
    """Find the intersection of two pd.DataFrames of time periods.

    Each row of each pd.DataFrame represents a single time period.  Each pd.DataFrame has
    two columns: `start_dt` and `end_dt` (where 'dt' is short for 'datetime').

    A typical use-case is that each pd.DataFrame represents all the time periods where
    a `DataSource` has contiguous, valid data.

    Here's a graphical example of two pd.DataFrames of time periods and their intersection:

                 ----------------------> TIME ->---------------------
               a: |-----|   |----|     |----------|     |-----------|
               b:    |--------|                       |----|    |---|
    intersection:    |--|   |-|                         |--|    |---|

    Args:
        a, b: pd.DataFrame where each row represents a time period.  The pd.DataFrame has
        two columns: start_dt and end_dt.

    Returns:
        Sorted list of intersecting time periods represented as a pd.DataFrame with two columns:
        start_dt and end_dt.
    """
    if a.empty or b.empty:
        return pd.DataFrame(columns=["start_dt", "end_dt"])

    all_intersecting_periods = []
    for a_period in a.itertuples():
        # Five ways in which two periods may overlap:
        # a: |----| or |---|   or  |---| or   |--|   or |-|
        # b:  |--|       |---|   |---|      |------|    |-|
        # In all five, `a` must always start before `b` ends,
        # and `a` must always end after `b` starts:
        overlapping_periods = b[(a_period.start_dt < b.end_dt) & (a_period.end_dt > b.start_dt)]

        # There are two ways in which two periods may *not* overlap:
        # a: |---|        or        |---|
        # b:       |---|      |---|
        # `overlapping` will not include periods which do *not* overlap.

        # Now find the intersection of each period in `overlapping_periods` with
        # the period from `a` that starts at `a_start_dt` and ends at `a_end_dt`.
        # We do this by clipping each row of `overlapping_periods`
        # to start no earlier than `a_start_dt`, and end no later than `a_end_dt`.

        # First, make a copy, so we don't clip the underlying data in `b`.
        intersecting_periods = overlapping_periods.copy()
        intersecting_periods.start_dt.clip(lower=a_period.start_dt, inplace=True)
        intersecting_periods.end_dt.clip(upper=a_period.end_dt, inplace=True)

        all_intersecting_periods.append(intersecting_periods)

    all_intersecting_periods = pd.concat(all_intersecting_periods)
    return all_intersecting_periods.sort_values(by="start_dt").reset_index(drop=True)


# From nowcasting_dataset.time:
def single_period_to_datetime_index(period: pd.Series, freq: str) -> pd.DatetimeIndex:
    """Return a DatetimeIndex from period['start_dt'] to period['end_dt'] at frequency freq.

    Before computing the date_range, this function first takes the ceiling of the
    start_dt (at frequency `freq`); and takes the floor of end_dt.  For example,
    if `freq` is '5 minutes' and `start_dt` is 12:03, then the ceiling of `start_dt`
    will be 12:05.  This is done so that all the returned datetimes are aligned to `freq`
    (e.g. if `freq` is '5 minutes' then every returned datetime will be at 00, 05, ..., 55
    minutes past the hour.
    """
    start_dt = period["start_dt"].ceil(freq)
    end_dt = period["end_dt"].floor(freq)
    return pd.date_range(start_dt, end_dt, freq=freq)


# From nowcasting_dataset.time:
def time_periods_to_datetime_index(time_periods: pd.DataFrame, freq: str) -> pd.DatetimeIndex:
    """Convert a DataFrame of time periods into a DatetimeIndex at a particular frequency.

    See the docstring of intersection_of_2_dataframes_of_periods() for more details.
    """
    assert len(time_periods) > 0
    dt_index = single_period_to_datetime_index(time_periods.iloc[0], freq=freq)
    for _, time_period in time_periods.iloc[1:].iterrows():
        new_dt_index = single_period_to_datetime_index(time_period, freq=freq)
        dt_index = dt_index.union(new_dt_index)
    return dt_index


def set_new_sample_period_and_t0_idx_attrs(xr_data, new_sample_period) -> xr.DataArray:
    orig_sample_period = xr_data.attrs["sample_period_duration"]
    orig_t0_idx = xr_data.attrs["t0_idx"]
    new_sample_period = pd.Timedelta(new_sample_period)
    assert new_sample_period >= orig_sample_period
    new_t0_idx = orig_t0_idx / (new_sample_period / orig_sample_period)
    np.testing.assert_almost_equal(
        int(new_t0_idx),
        new_t0_idx,
        err_msg=(
            "The original t0_idx must be exactly divisible by"
            " (new_sample_period / orig_sample_period)"
        ),
    )
    xr_data.attrs["sample_period_duration"] = new_sample_period
    xr_data.attrs["t0_idx"] = int(new_t0_idx)
    return xr_data



import glob
import os
from pathlib import Path

import numpy as np

from power_perceiver.load_prepared_batches.prepared_dataset import PreparedDataset
from power_perceiver.load_prepared_batches.data_sources import PV, GSP, HRVSatellite, NWP
from power_perceiver.production.model import FullModel
import pandas as pd
from torch.utils.data import DataLoader

dataset = PreparedDataset([PV(history_duration=pd.Timedelta("90 min"),),
                           GSP(history_duration=pd.Timedelta("2 hours"),),
                           HRVSatellite(history_duration=pd.Timedelta("30 min"),),
                           NWP(history_duration=pd.Timedelta("1 hour"),)],
                          data_path=Path("/home/jacob/Development/power_perceiver/data_for_testing/"))
print(dataset)
dataloader = DataLoader(dataset, num_workers=0, batch_size=None)
batch = next(iter(dataloader))
model = FullModel().eval()
model(batch)
for key, value in batch.items():
    if isinstance(value, np.ndarray):
        print(f"{key}: {value.shape}")
    else:
        print(f"{key}: {value}")




import datetime
import logging
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, ClassVar, Iterable, Optional, Union

import numpy as np
import pandas as pd
import xarray as xr

from power_perceiver.consts import Location
from power_perceiver.exceptions import NoPVSystemsInSlice
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import NumpyBatch
from power_perceiver.time import get_contiguous_time_periods
from power_perceiver.utils import check_path_exists

_log = logging.getLogger(__name__)


# `kw_only` allows a base class to have fields with default values,
# and it's still OK for child classes to have fields *without* defaults. See:
# https://stackoverflow.com/a/69822584/732596
@dataclass(kw_only=True)
class RawDataSource:
    """Abstract base class for loading data directly from the raw or intermediate files.

    Attributes:
        data_in_ram: xr.DataArray. Uses standard names for dimensions and coordinates. If the data
            does not conform to these standard names, then the subclass must change the names
            as soon as the DataArray is opened. The standard names are:
                Dimension names: time_utc, channel
                Coordinate names: time_utc, y_osbg, x_osgb, y_geostationary, x_geostationary
        data_on_disk: xr.DataArray.
            If this DataSource loads a subset of the full DataArray at the start of each epoch,
            then `data_in_ram` holds the in-memory data, and `data_on_disk` holds
            the un-loaded `xr.DataArray`. If this DataSource can load the entire DataArray into RAM,
            then `data_on_disk` will be `None`.
    """

    transforms: Optional[Iterable[Callable]] = None
    needs_to_load_subset_into_ram: ClassVar[bool] = False

    def __post_init__(self):  # noqa: D105
        self._data_in_ram = None
        self._data_on_disk = None
        # For data sources that can load everything into RAM,
        # override `__post_init__` to load everything into RAM here.
        # Otherwise, override `per_worker_init` to load everything into RAM there.

    @property
    def data_in_ram(self):  # noqa: D102
        if self._data_in_ram is None:
            raise RuntimeError("Please load the data into RAM before accessing `data_in_ram`!")
        return self._data_in_ram

    @property
    def data_on_disk(self):  # noqa: D102
        if self._data_on_disk is None:
            raise RuntimeError("Please open the dataset before accessing `data_on_disk`!")
        return self._data_on_disk

    def per_worker_init(self, worker_id: int, seed: int) -> None:  # noqa: D102
        self.worker_id = worker_id
        # Each worker must have a different seed for its random number generator.
        # Otherwise all the workers will output exactly the same data!
        _log.info(f"{worker_id=} has random number generator {seed=:,d}")
        self.rng = np.random.default_rng(seed=seed)
        # Optionally, override `per_worker_init` in child class to call super().per_worker_init()
        # and to `open()` the data, if opening doesn't happen in `__post_init__().`

    def get_example(
        self, t0_datetime_utc: datetime.datetime, center_osgb: Location
    ) -> xr.DataArray:
        """Can be overridden by child classes.

        The returned Dataset must not include an `example` dimension.
        """
        using_empty_example = False
        try:
            xr_data = self._get_slice(t0_datetime_utc=t0_datetime_utc, center_osgb=center_osgb)
        except NoPVSystemsInSlice:
            xr_data = self.empty_example
            using_empty_example = True
        xr_data = self._post_process(xr_data)
        xr_data = self._set_attributes(xr_data)
        xr_data = self._transform(xr_data)

        if not using_empty_example:
            try:
                self.check_xarray_data(xr_data)
            except Exception as e:
                raise e.__class__(
                    f"Exception raised when checking xr data! {t0_datetime_utc=} {center_osgb=}"
                ) from e
        return xr_data

    def _get_slice(self, t0_datetime_utc: datetime.datetime, center_osgb: Location) -> xr.DataArray:
        """Can be overridden by child classes.

        The returned Dataset must not include an `example` dimension.
        """
        xr_data = self.data_in_ram
        xr_data = self._get_time_slice(xr_data, t0_datetime_utc=t0_datetime_utc)
        xr_data = self._get_spatial_slice(xr_data, center_osgb=center_osgb)
        return xr_data

    def check_xarray_data(self, xr_data: xr.DataArray):  # noqa: D102
        assert np.isfinite(xr_data).all(), "Some xr_data is non-finite!"

    def _set_attributes(self, xr_data: xr.DataArray) -> xr.DataArray:
        return xr_data

    @staticmethod
    def to_numpy(xr_data: xr.DataArray) -> NumpyBatch:
        """Convert xarray to numpy batch.

        But note that this is actually just returns *one* example (not a whole batch!)
        """
        raise NotImplementedError("Must be implemented by subclass!")

    def _get_time_slice(
        self, xr_data: xr.DataArray, t0_datetime_utc: datetime.datetime
    ) -> xr.DataArray:
        """Can be overridden, usually by TimeseriesDataSource.

        The returned Dataset does not include an `example` dimension.

        This method should sanity check that the example is the correct duration. But shouldn't
        check for NaNs (because subsequent processing might remove NaNs.
        `check_xarray_data` checks for NaNs.
        """
        return xr_data

    def _get_spatial_slice(self, xr_data: xr.DataArray, center_osgb: Location) -> xr.DataArray:
        """Can be overridden, usually by SpatialDataSource.

        The returned Dataset does not include an `example` dimension.

        This method should sanity check that the example is the correct spatial shape. But shouldn't
        check for NaNs (because subsequent processing might remove NaNs.
        `check_xarray_data` checks for NaNs.
        """
        return xr_data

    def _post_process(self, xr_data: xr.DataArray) -> xr.DataArray:
        """Can be overridden. This is where normalisation should happen.

        The returned Dataset does not include an `example` dimension.
        """
        return xr_data

    def _transform(self, xr_data: xr.DataArray) -> xr.DataArray:
        if self.transforms:
            for transform in self.transforms:
                xr_data = transform(xr_data)
        return xr_data


@dataclass(kw_only=True)
class TimeseriesDataSource:
    """Abstract base class mixin for timeseries data sources.

    Init args:
        history_duration: Total duration of the history *including* t0.
            (This is a different definition to the definition used in `nowcasting_dataset`,
            which *excludes* t0 from `history_minutes`.) `history_duration` is the duration
            between the first timestep and t0. We consider t0 to be part of the history.
            `history_duration` can be zero. If `history_duration` and `forecast_duration` are both
            zero then there will still be a single timestep at t0.
        forecast_duration: Total duration of the forecast: dt_end - t0.
        sample_period_duration: The maximum legal `timedelta` between samples. Cannot be zero.
        time_periods: The time periods to consider. A pd.DataFrame with two columns:
            start_dt and end_dt.
    """

    history_duration: datetime.timedelta
    forecast_duration: datetime.timedelta
    time_periods: pd.DataFrame
    sample_period_duration: ClassVar[datetime.timedelta]
    _time_dim_name: ClassVar[str] = "time_utc"

    def __post_init__(self):  # noqa: D105
        # Sanity checks.
        assert self.sample_period_duration, "sample_period_duration cannot be zero length!"
        assert (self.time_periods["start_dt"] < self.time_periods["end_dt"]).all()
        start_dt = pd.DatetimeIndex(self.time_periods["start_dt"])
        assert start_dt.is_monotonic_increasing
        if self.history_duration:
            assert (
                self.history_duration >= self.sample_period_duration
            ), f"{self.history_duration=} must be zero or >= {self.sample_period_duration}"

        if self.forecast_duration:
            assert (
                self.forecast_duration >= self.sample_period_duration
            ), f"{self.forecast_duration=} must be zero or >= {self.sample_period_duration}"

    @property
    def total_seq_length(self) -> int:
        """Total number of timesteps per example, including t0.

        Our definition "total_seq_length" is inclusive of both the first and last timestep.
        """
        # Plus 1 because "total_seq_length" is inclusive of both the first and last timestep.
        return int(self.total_duration / self.sample_period_duration) + 1

    @property
    def total_duration(self) -> datetime.timedelta:
        """Total number of duration per example, including t0."""
        return self.history_duration + self.forecast_duration

    @property
    def t0_idx(self) -> int:
        """The index into the array for the most recent observation (t0).

        Remember that, in this code, we consider t0 to be part of the history.
        So, if the history_duration is 1 hour, and sample_period_duration is 30 minutes,
        then "history" will be at indicies 0, 1, and 2.
        """
        return int(self.history_duration / self.sample_period_duration)

    @property
    def datetime_index(self) -> pd.DatetimeIndex:
        """Return a pd.DatetimeIndex of all available datetimes.

        Child classes can override this to, for example, filter out any datetimes
        which don't make sense for this DataSource, e.g. remove nighttime.
        """
        if self.needs_to_load_subset_into_ram:
            data = self.data_on_disk
        else:
            data = self.data_in_ram
        return pd.DatetimeIndex(data[self._time_dim_name])

    def load_subset_into_ram(self, subset_of_contiguous_t0_time_periods: pd.DataFrame) -> None:
        """Load a subset of `data_on_disk` into `data_in_ram`.

        Args:
            subset_of_contiguous_t0_time_periods: DataFrame with columns 'start_dt' and 'end_dt'
                specifying the start and end of value t0 periods.

        Override in DataSources which can only fit a subset of the dataset into RAM.
        """
        _log.info(f"{self.__class__.__name__} load_subset_into_ram().")
        # Delete any existing data in RAM while we're loading new data.
        self._data_in_ram = None

        # Convert t0_time_periods back into the complete time periods we want to load:
        time_periods = self._convert_t0_time_periods_to_periods_to_load(
            subset_of_contiguous_t0_time_periods
        )
        assert (time_periods["start_dt"] < time_periods["end_dt"]).all()

        # Lazily create a new DataArray with just the data we want.
        data_to_load = []
        for _, row in time_periods.iterrows():
            start_dt = row["start_dt"]
            end_dt = row["end_dt"]
            data_for_period = self.data_on_disk.sel({self._time_dim_name: slice(start_dt, end_dt)})
            data_to_load.append(data_for_period)

        # Load into RAM :)
        data_to_load = xr.concat(data_to_load, dim=self._time_dim_name)
        data_to_load = data_to_load.drop_duplicates(dim=self._time_dim_name)
        self._data_in_ram = data_to_load.load()

    def _convert_t0_time_periods_to_periods_to_load(
        self, subset_of_contiguous_t0_time_periods: pd.DataFrame
    ) -> pd.DataFrame:
        """Convert t0_time_periods back into the complete time periods we want to load."""
        time_periods = deepcopy(subset_of_contiguous_t0_time_periods)
        del subset_of_contiguous_t0_time_periods
        time_periods["start_dt"] -= self.history_duration
        time_periods["end_dt"] += self.forecast_duration
        return time_periods

    def get_contiguous_t0_time_periods(self) -> pd.DataFrame:
        """Get all time periods which contain valid t0 datetimes.

        `t0` is the datetime of the most recent observation.

        Returns:
          pd.DataFrame where each row represents a single time period.  The pd.DataFrame
          has two columns: `start_dt` and `end_dt` (where 'dt' is short for 'datetime').
        """
        contiguous_time_periods = self._get_contiguous_time_periods()
        contiguous_time_periods["start_dt"] += self.history_duration
        contiguous_time_periods["end_dt"] -= self.forecast_duration
        assert (contiguous_time_periods["start_dt"] < contiguous_time_periods["end_dt"]).all()
        return contiguous_time_periods

    def _get_contiguous_time_periods(self) -> pd.DataFrame:
        """Get all the time periods for which this DataSource has contiguous data.

        Returns:
          pd.DataFrame where each row represents a single time period.  The pd.DataFrame
          has two columns: `start_dt` and `end_dt` (where 'dt' is short for 'datetime').
        """
        return get_contiguous_time_periods(
            datetimes=self.datetime_index,
            min_seq_length=self.total_seq_length,
            max_gap_duration=self.sample_period_duration,
        )

    def _get_time_slice(
        self, xr_data: xr.DataArray, t0_datetime_utc: datetime.datetime
    ) -> xr.DataArray:
        """Select a timeslice from `xr_data`.

        The returned data does not include an `example` dimension.
        """
        start_dt_ceil = self._get_start_dt_ceil(t0_datetime_utc)
        end_dt_ceil = self._get_end_dt_ceil(t0_datetime_utc)

        # Sanity check!
        assert (
            start_dt_ceil in xr_data.time_utc
        ), f"{start_dt_ceil=} not in xr_data.time_utc! {t0_datetime_utc=}"
        assert (
            end_dt_ceil in xr_data.time_utc
        ), f"{end_dt_ceil=} not in xr_data.time_utc! {t0_datetime_utc=}"

        # Get time slice:
        time_slice = xr_data.sel({self._time_dim_name: slice(start_dt_ceil, end_dt_ceil)})
        self._sanity_check_time_slice(time_slice, self._time_dim_name, t0_datetime_utc)
        return time_slice

    def _sanity_check_time_slice(
        self,
        time_slice: xr.DataArray,
        time_dim_name: str,
        t0_datetime_utc: datetime.datetime,
    ) -> None:
        info_str = f"Context: {t0_datetime_utc=}; {time_dim_name=}; {time_slice=}"
        assert (
            len(time_slice[time_dim_name]) == self.total_seq_length
        ), f"{len(time_slice[time_dim_name])=} != {self.total_seq_length=} {info_str}"
        time_slice_duration = np.timedelta64(
            time_slice[time_dim_name][-1].values - time_slice[time_dim_name][0].values
        )
        expected_duration = np.timedelta64(self.total_duration)
        assert (
            time_slice_duration == expected_duration
        ), f"{time_slice_duration=} != {expected_duration=} {info_str}"

    def _get_start_dt_ceil(
        self, t0_datetime_utc: Union[datetime.datetime, np.datetime64, pd.Timestamp]
    ) -> pd.Timestamp:
        start_dt = pd.Timestamp(t0_datetime_utc) - np.timedelta64(self.history_duration)
        return start_dt.ceil(self.sample_period_duration)

    def _get_end_dt_ceil(
        self, t0_datetime_utc: Union[datetime.datetime, np.datetime64, pd.Timestamp]
    ) -> pd.Timestamp:
        end_dt = pd.Timestamp(t0_datetime_utc) + np.timedelta64(self.forecast_duration)
        return end_dt.ceil(self.sample_period_duration)

    def _set_attributes(self, xr_data: xr.DataArray) -> xr.DataArray:
        xr_data.attrs["t0_idx"] = self.t0_idx
        xr_data.attrs["sample_period_duration"] = self.sample_period_duration
        return xr_data


@dataclass(kw_only=True)
class SpatialDataSource:
    """Abstract base class for dense spatial image data sources (like NWP and satellite).

    Sparse spatial data (such as PV data) doesn't inherit from `SpatialDataSource`.

    Args:
        roi_height_pixels: Height of the image in each example. Must be divisible by 2.
            ROI stands for region of interest.
        roi_width_pixels: Width of the image in each example. Must be divisible by 2.
    """

    roi_height_pixels: int
    roi_width_pixels: int

    # Attributes which are intended to be set for the whole class.
    _y_dim_name: ClassVar[str] = "y"
    _x_dim_name: ClassVar[str] = "x"

    def __post_init__(self):
        assert self.roi_height_pixels > 0, f"{self.roi_height_pixels=} must be > 0!"
        assert self.roi_width_pixels > 0, f"{self.roi_width_pixels=} must be > 0!"
        assert (
            self.roi_height_pixels % 2
        ) == 0, f"{self.roi_height_pixels=} must be divisible by 2!"
        assert (self.roi_width_pixels % 2) == 0, f"{self.roi_width_pixels=} must be divisible by 2!"

    def get_osgb_location_for_example(self) -> Location:
        """Randomly select a valid geographical location for one example.

        Returns:  Location(x_center_osgb, y_center_osgb)
        """
        # Find the minimum and maximum legal values for the randomly sampled x and y positions:
        half_height_of_crop = self.roi_height_pixels // 2
        half_width_of_crop = self.roi_width_pixels // 2

        source_image_height = len(self._data_in_ram[self._y_dim_name])
        source_image_width = len(self._data_in_ram[self._x_dim_name])

        # Plus or minus one for safety.
        min_y = half_height_of_crop + 1
        max_y = source_image_height - half_height_of_crop - 1
        min_x = half_width_of_crop + 1
        max_x = source_image_width - half_width_of_crop - 1

        # Sanity check!
        assert 0 < min_x < source_image_width
        assert 0 < max_x < source_image_width
        assert 0 < min_y < source_image_height
        assert 0 < max_y < source_image_height

        # Randomly pick y and x indexes.
        y_idx = self.rng.integers(low=min_y, high=max_y)
        x_idx = self.rng.integers(low=min_x, high=max_x)

        # Get the OSGB coords for those indexes:
        selected_pixel = self._data_in_ram.isel(
            {
                self._y_dim_name: y_idx,
                self._x_dim_name: x_idx,
            }
        )
        y_osgb = selected_pixel.y_osgb.item()
        x_osgb = selected_pixel.x_osgb.item()

        return Location(x=x_osgb, y=y_osgb)

    def _get_spatial_slice(self, xr_data: xr.DataArray, center_osgb: Location) -> xr.DataArray:
        """Slice `xr_data` to produce a region of interest, centered on `center_osgb`.

        Assume the image data starts top-left.

        The returned data does not include an `example` dimension.
        """
        # Find pixel index at `center_osgb`:
        center_idx = self._get_idx_of_pixel_at_center_of_roi(
            xr_data=xr_data, center_osgb=center_osgb
        )

        # Compute the index for left and right:
        half_height = self.roi_height_pixels // 2
        half_width = self.roi_width_pixels // 2

        left_idx = center_idx.x - half_width
        right_idx = center_idx.x + half_width
        top_idx = center_idx.y - half_height
        bottom_idx = center_idx.y + half_height

        # Sanity check!
        assert left_idx >= 0, f"{left_idx=} must be >= 0!"
        data_width_pixels = len(xr_data[self._x_dim_name])
        assert right_idx <= data_width_pixels, f"{right_idx=} must be <= {data_width_pixels=}"
        assert top_idx >= 0, f"{top_idx=} must be >= 0!"
        data_height_pixels = len(xr_data[self._y_dim_name])
        assert bottom_idx <= data_height_pixels, f"{bottom_idx=} must be <= {data_height_pixels=}"

        selected = xr_data.isel(
            {
                self._x_dim_name: slice(left_idx, right_idx),
                self._y_dim_name: slice(top_idx, bottom_idx),
            }
        )

        # Sanity check:
        assert len(selected[self._x_dim_name]) == self.roi_width_pixels
        assert len(selected[self._y_dim_name]) == self.roi_height_pixels

        return selected

    def _get_idx_of_pixel_at_center_of_roi(
        self, xr_data: xr.DataArray, center_osgb: Location
    ) -> Location:
        """Return x and y index location of pixel at center of region of interest."""
        y_index = xr_data.get_index(self._y_dim_name)
        x_index = xr_data.get_index(self._x_dim_name)
        return Location(
            y=y_index.get_indexer([center_osgb.y], method="nearest")[0],
            x=x_index.get_indexer([center_osgb.x], method="nearest")[0],
        )


@dataclass(kw_only=True)
class ZarrDataSource:
    """Abstract base class for Zarr stores."""

    zarr_path: Union[Path, str]

    def __post_init__(self):
        check_path_exists(self.zarr_path)



"""
U-Net followed by the Power Perceiver.

See this issue for a diagram: https://github.com/openclimatefix/power_perceiver/issues/54
"""

# General imports
import datetime
import logging
import socket
from dataclasses import dataclass
from typing import Optional

import einops

# ML imports
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from pytorch_lightning.loggers import WandbLogger
from pytorch_msssim import ms_ssim
from torch import nn

from power_perceiver.analysis.log_national_pv import LogNationalPV
from power_perceiver.analysis.plot_probability_timeseries import LogProbabilityTimeseriesPlots
from power_perceiver.analysis.plot_tsne import LogTSNEPlot

# power_perceiver imports
from power_perceiver.consts import X_OSGB_MEAN, X_OSGB_STD, Y_OSGB_MEAN, Y_OSGB_STD, BatchKey
from power_perceiver.load_prepared_batches.data_sources.satellite import SAT_MEAN, SAT_STD
from power_perceiver.load_raw.data_sources.raw_gsp_data_source import RawGSPDataSource
from power_perceiver.load_raw.data_sources.raw_nwp_data_source import RawNWPDataSource
from power_perceiver.load_raw.data_sources.raw_pv_data_source import RawPVDataSource
from power_perceiver.load_raw.data_sources.raw_satellite_data_source import RawSatelliteDataSource
from power_perceiver.load_raw.national_pv_dataset import NationalPVDataset
from power_perceiver.load_raw.raw_dataset import RawDataset
from power_perceiver.np_batch_processor.align_gsp_to_5_min import AlignGSPTo5Min
from power_perceiver.np_batch_processor.delete_forecast_satellite_imagery import (
    DeleteForecastSatelliteImagery,
)
from power_perceiver.np_batch_processor.encode_space_time import EncodeSpaceTime
from power_perceiver.np_batch_processor.save_t0_time import SaveT0Time
from power_perceiver.np_batch_processor.sun_position import SunPosition
from power_perceiver.np_batch_processor.topography import Topography
from power_perceiver.pytorch_modules.mixture_density_network import (
    MixtureDensityNetwork,
    get_distribution,
)
from power_perceiver.pytorch_modules.nwp_processor import NWPProcessor
from power_perceiver.pytorch_modules.query_generator import GSPQueryGenerator, PVQueryGenerator
from power_perceiver.pytorch_modules.satellite_predictor import XResUNet
from power_perceiver.pytorch_modules.satellite_processor import HRVSatelliteProcessor
from power_perceiver.pytorch_modules.self_attention import MultiLayerTransformerEncoder
from power_perceiver.transforms.pv import PVPowerRollingWindow
from power_perceiver.utils import assert_num_dims, pandas_periods_to_our_periods_dt
from power_perceiver.xr_batch_processor.reduce_num_timesteps import random_int_without_replacement

logging.basicConfig()
_log = logging.getLogger("power_perceiver")
_log.setLevel(logging.DEBUG)

plt.rcParams["figure.figsize"] = (18, 10)
plt.rcParams["figure.facecolor"] = "white"

# SatellitePredictor options
NUM_HIST_SAT_IMAGES = 12  # v15 pre-prepared batches use 7
NUM_FUTURE_SAT_IMAGES = 25  # v15 pre-prepared batches use 24
USE_TOPOGRAPHY = True
USE_SUN_POSITION = True

SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS = 256
SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS = 128

SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS = 64
N_PV_SYSTEMS_PER_EXAMPLE = 8

# PowerPerceiver options
D_MODEL = 256  # Must be divisible by N_HEADS
N_HEADS = 32

NWP_CHANNELS = ("t", "dswrf", "prate", "r", "si10", "vis", "lcc", "mcc", "hcc")

ON_DONATELLO = socket.gethostname() == "donatello"

DEBUG = False
ENABLE_WANDB = True

if DEBUG:
    GPUS = [0]
elif ON_DONATELLO:
    GPUS = [0, 1, 2, 4]
else:  # On GCP
    GPUS = [0, 1]


def get_dataloader(
    time_periods,
    num_workers,
    n_batches_per_epoch_per_worker,
    load_subset_every_epoch,
    train: bool,
) -> torch.utils.data.DataLoader:

    data_source_kwargs = dict(
        time_periods=time_periods,
        history_duration=datetime.timedelta(hours=1),
        forecast_duration=datetime.timedelta(hours=2),
    )

    sat_data_source = RawSatelliteDataSource(
        zarr_path=(
            (
                "/mnt/storage_ssd_4tb/data/ocf/solar_pv_nowcasting/nowcasting_dataset_pipeline/"
                "satellite/EUMETSAT/SEVIRI_RSS/zarr/v3/eumetsat_seviri_hrv_uk.zarr"
            )
            if ON_DONATELLO
            else (
                "gs://solar-pv-nowcasting-data/"
                "satellite/EUMETSAT/SEVIRI_RSS/v3/eumetsat_seviri_hrv_uk.zarr"
            )
        ),
        roi_height_pixels=SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS,
        roi_width_pixels=SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS,
        **data_source_kwargs,
    )

    pv_data_source = RawPVDataSource(
        pv_power_filename="/home/jack/data/PV/Passiv/ocf_formatted/v0/passiv.netcdf",
        pv_metadata_filename="/home/jack/data/PV/Passiv/ocf_formatted/v0/system_metadata_OCF_ONLY.csv",
        roi_height_meters=96_000,
        roi_width_meters=96_000,
        n_pv_systems_per_example=N_PV_SYSTEMS_PER_EXAMPLE,
        transforms=[PVPowerRollingWindow(expect_dataset=False)],
        **data_source_kwargs,
    )

    gsp_data_source = RawGSPDataSource(
        gsp_pv_power_zarr_path="/home/jack/data/PV/GSP/v3/pv_gsp.zarr",
        gsp_id_to_region_id_filename="/home/jack/data/PV/GSP/eso_metadata.csv",
        sheffield_solar_region_path="/home/jack/data/PV/GSP/gsp_shape",
        time_periods=time_periods,
        history_duration=datetime.timedelta(hours=1),
        forecast_duration=datetime.timedelta(hours=8),
    )

    nwp_data_source = RawNWPDataSource(
        zarr_path=(
            (
                "/mnt/storage_ssd_4tb/data/ocf/solar_pv_nowcasting/nowcasting_dataset_pipeline/"
                "NWP/UK_Met_Office/UKV/zarr/UKV_intermediate_version_3.zarr"
            )
            if ON_DONATELLO
            else ("gs://solar-pv-nowcasting-data/NWP/UK_Met_Office/UKV_intermediate_version_3.zarr")
        ),
        roi_height_pixels=4,
        roi_width_pixels=4,
        y_coarsen=16,
        x_coarsen=16,
        channels=NWP_CHANNELS,
        time_periods=time_periods,
        history_duration=datetime.timedelta(hours=1),
        forecast_duration=datetime.timedelta(hours=8),
    )

    np_batch_processors = [AlignGSPTo5Min(), EncodeSpaceTime(), SaveT0Time()]
    if USE_SUN_POSITION:
        for modality_name in ["hrvsatellite", "gsp", "gsp_5_min", "pv", "nwp_target_time"]:
            np_batch_processors.append(SunPosition(modality_name=modality_name))
    if USE_TOPOGRAPHY:
        np_batch_processors.append(Topography("/home/jack/europe_dem_2km_osgb.tif"))
    # Delete imagery of the future, because we're not training the U-Net,
    # and we want to save GPU RAM.
    # But we do want hrvsatellite_time_utc to continue into the future by 2 hours because
    # downstream code relies on hrvsatellite_time_utc.
    # This must come last.
    np_batch_processors.append(
        DeleteForecastSatelliteImagery(num_hist_sat_images=NUM_HIST_SAT_IMAGES)
    )

    raw_dataset_kwargs = dict(
        n_examples_per_batch=14,  # TODO: Increase to more like 32!
        n_batches_per_epoch=n_batches_per_epoch_per_worker,
        np_batch_processors=np_batch_processors,
        load_subset_every_epoch=load_subset_every_epoch,
        min_duration_to_load_per_epoch=datetime.timedelta(
            hours=48 if DEBUG else ((12 * 32) if ON_DONATELLO else (12 * 24))
        ),
        data_source_combos=dict(
            gsp_pv_nwp_sat=(gsp_data_source, pv_data_source, nwp_data_source, sat_data_source),
        ),
    )

    if train:
        raw_dataset = RawDataset(**raw_dataset_kwargs)
    else:
        raw_dataset = NationalPVDataset(**raw_dataset_kwargs)

    if not num_workers:
        raw_dataset.per_worker_init(worker_id=0)

    def _worker_init_fn(worker_id: int):
        worker_info = torch.utils.data.get_worker_info()
        dataset_obj = worker_info.dataset
        if len(GPUS) > 1:
            rank = torch.distributed.get_rank()
        else:
            rank = 0
        _log.info(f"{num_workers=} {worker_id=} {rank=} {worker_info.seed=}")
        # We set the worker_id to be unique across all GPUs, so each worker
        # sets a unique (but repeatable) random number generator seed.
        # We times by 16 just to really make sure that the random seeds
        # are different for each GPU.
        seed = worker_info.seed + worker_id + (rank * num_workers * 16)
        dataset_obj.per_worker_init(worker_id=worker_id, seed=seed)

    dataloader = torch.utils.data.DataLoader(
        raw_dataset,
        batch_size=None,
        num_workers=num_workers,
        pin_memory=True,
        worker_init_fn=_worker_init_fn,
        persistent_workers=num_workers > 0,
    )

    return dataloader


# ---------------------------------- SatellitePredictor ----------------------------------


def get_osgb_coords_for_coord_conv(batch: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
    """Returns tensor of shape (example, 2, y, x)."""
    y_osgb = batch[BatchKey.hrvsatellite_y_osgb]
    x_osgb = batch[BatchKey.hrvsatellite_x_osgb]

    # Normalise:
    y_osgb = (y_osgb - Y_OSGB_MEAN) / Y_OSGB_STD
    x_osgb = (x_osgb - X_OSGB_MEAN) / X_OSGB_STD

    # Concat:
    return torch.stack((y_osgb, x_osgb), dim=1)


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class SatellitePredictor(pl.LightningModule):
    use_coord_conv: bool = False
    crop: bool = False
    use_topography: bool = USE_TOPOGRAPHY
    use_sun_position: bool = USE_SUN_POSITION

    # kwargs to fastai DynamicUnet. See this page for details:
    # https://fastai1.fast.ai/vision.models.unet.html#DynamicUnet
    pretrained: bool = False
    blur_final: bool = True  # Blur final layer. fastai default is True.
    self_attention: bool = True  # Use SA layer at the third block before the end.
    last_cross: bool = True  # Use a cross-connection with the direct input of the model.
    bottle: bool = False  # Bottleneck the last skip connection.

    def __post_init__(self):
        super().__init__()

        self.satellite_predictor = XResUNet(
            img_size=(
                SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS,
                SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS,
            ),
            n_in=(
                NUM_HIST_SAT_IMAGES
                + (2 if self.use_coord_conv else 0)
                + (1 if self.use_topography else 0)
                + (2 if self.use_sun_position else 0)
            ),
            n_out=NUM_FUTURE_SAT_IMAGES,
            pretrained=self.pretrained,
            blur_final=self.blur_final,
            self_attention=self.self_attention,
            last_cross=self.last_cross,
            bottle=self.bottle,
        )

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
        data = x[BatchKey.hrvsatellite_actual][
            :, :NUM_HIST_SAT_IMAGES, 0
        ]  # Shape: (example, time, y, x)
        height, width = data.shape[2:]
        assert height == SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS, f"{height=}"
        assert width == SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS, f"{width=}"

        if self.use_coord_conv:
            osgb_coords = get_osgb_coords_for_coord_conv(x)
            data = torch.concat((data, osgb_coords), dim=1)

        if self.use_topography:
            surface_height = x[BatchKey.hrvsatellite_surface_height]
            surface_height = surface_height.unsqueeze(1)  # Add channel dim
            data = torch.concat((data, surface_height), dim=1)

        if self.use_sun_position:
            azimuth_at_t0 = x[BatchKey.hrvsatellite_solar_azimuth][:, NUM_HIST_SAT_IMAGES]
            elevation_at_t0 = x[BatchKey.hrvsatellite_solar_elevation][:, NUM_HIST_SAT_IMAGES]
            sun_pos = torch.stack((azimuth_at_t0, elevation_at_t0), dim=1)  # Shape: (example, 2)
            del azimuth_at_t0, elevation_at_t0
            # Repeat over y and x:
            sun_pos = einops.repeat(sun_pos, "example chan -> example chan y x", y=height, x=width)
            data = torch.concat((data, sun_pos), dim=1)

        assert data.isfinite().all()
        predicted_sat = self.satellite_predictor(data)
        assert predicted_sat.isfinite().all()
        return predicted_sat  # Shape: example, time, y, x

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        predicted_sat = self(batch)
        actual_sat = batch[BatchKey.hrvsatellite_actual][:, NUM_HIST_SAT_IMAGES:, 0]

        sat_mse_loss = F.mse_loss(predicted_sat, actual_sat)
        self.log(f"{self.tag}/sat_mse", sat_mse_loss)

        # MS-SSIM. Requires images to be de-normalised:
        actual_sat_denorm = (actual_sat * SAT_STD["HRV"]) + SAT_MEAN["HRV"]
        predicted_sat_denorm = (predicted_sat * SAT_STD["HRV"]) + SAT_MEAN["HRV"]
        ms_ssim_loss = 1 - ms_ssim(
            predicted_sat_denorm,
            actual_sat_denorm,
            data_range=1023.0,
            size_average=True,  # Return a scalar.
            win_size=3,  # ClimateHack folks used win_size=3.
        )
        self.log(f"{self.tag}/ms_ssim", ms_ssim_loss)
        self.log(f"{self.tag}/ms_ssim+sat_mse", ms_ssim_loss + sat_mse_loss)

        if self.crop_sat_before_sat_predictor_loss:
            # Loss on a central crop:
            # The cropped image has to be larger than 32x32 otherwise ms-ssim complains:
            # "Image size should be larger than 32 due to the 4 downsamplings in ms-ssim"
            sat_mse_loss_crop = F.mse_loss(
                predicted_sat[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
                actual_sat[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
            )
            self.log(f"{self.tag}/sat_mse_crop", sat_mse_loss_crop)
            ms_ssim_loss_crop = 1 - ms_ssim(
                predicted_sat_denorm[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
                actual_sat_denorm[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
                data_range=1023.0,
                size_average=True,  # Return a scalar.
                win_size=3,  # The Illinois ClimateHack folks used win_size=3.
            )
            self.log(f"{self.tag}/ms_ssim_crop", ms_ssim_loss_crop)
            self.log(f"{self.tag}/ms_ssim_crop+sat_mse_crop", ms_ssim_loss_crop + sat_mse_loss_crop)
            sat_loss = ms_ssim_loss_crop + sat_mse_loss_crop
        else:
            sat_loss = ms_ssim_loss + sat_mse_loss

        return dict(
            lost=sat_loss,
        )


# ---------------------------------- SatelliteTransformer ----------------------------------


def maybe_pad_with_zeros(tensor: torch.Tensor, requested_dim: int) -> torch.Tensor:
    num_zeros_to_pad = requested_dim - tensor.shape[-1]
    assert num_zeros_to_pad >= 0, f"{requested_dim=}, {tensor.shape=}"
    if num_zeros_to_pad > 0:
        zero_padding_shape = tensor.shape[:2] + (num_zeros_to_pad,)
        zero_padding = torch.zeros(*zero_padding_shape, dtype=tensor.dtype, device=tensor.device)
        tensor = torch.concat((tensor, zero_padding), dim=2)
    return tensor


# Indexes for cropping the centre of the satellite image:
LEFT_IDX = (SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS // 2) - (
    SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS // 2
)
RIGHT_IDX = LEFT_IDX + SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS
TOP_IDX = (SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS // 2) - (
    SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS // 2
)
BOTTOM_IDX = TOP_IDX + SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS


def _crop_satellite_and_spatial_coords(
    x: dict[BatchKey, torch.Tensor]
) -> dict[BatchKey, torch.Tensor]:
    """Crop satellite data & spatial coordinates.

    This is necessary because we want to give the "satellite predictor"
    a large rectangle of imagery (e.g. 256 wide x 128 high) so it can see clouds
    coming, but we probably don't want to give that huge image to a full-self-attention model.
    """
    assert_num_dims(x[BatchKey.hrvsatellite_actual], num_expected_dims=5)
    assert_num_dims(x[BatchKey.hrvsatellite_predicted], num_expected_dims=4)

    def _check_shape(batch_key, y_idx=1, x_idx=2):
        error_msg = f"{batch_key.name}.shape = {x[batch_key].shape}"
        assert x[batch_key].shape[y_idx] == SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS, error_msg
        assert x[batch_key].shape[x_idx] == SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS, error_msg

    # Crop the predicted and actual imagery:
    for batch_key in (BatchKey.hrvsatellite_actual, BatchKey.hrvsatellite_predicted):
        x[batch_key] = x[batch_key][..., TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX]
        _check_shape(batch_key, y_idx=-2, x_idx=-1)

    # Crop the coords (which have to be done in a separate loop, because `y` and `x`
    # are at different positions):
    for batch_key in (
        BatchKey.hrvsatellite_y_osgb_fourier,
        BatchKey.hrvsatellite_x_osgb_fourier,
        BatchKey.hrvsatellite_surface_height,
    ):
        x[batch_key] = x[batch_key][:, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX]
        _check_shape(batch_key)
    return x


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class SatelliteTransformer(nn.Module):
    """Infers a single timestep of PV power and GSP power at a time.

    Currently just uses HRV satellite imagery as the input. In the near future it could also
    use NWP temperature, wind speed & precipitation, and absolute geo position.
    """

    # Params for Perceiver
    # byte_array and query will be automatically padded with zeros to get to d_model.
    # Set d_model to be divisible by `num_heads`.
    d_model: int = D_MODEL
    pv_system_id_embedding_dim: int = 16
    num_heads: int = N_HEADS
    dropout: float = 0.0
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 8

    def __post_init__(self):
        super().__init__()
        self.hrvsatellite_processor = HRVSatelliteProcessor()

        NUM_PV_SYSTEMS = 1400
        NUM_GSPS = 360
        id_embedding = nn.Embedding(
            num_embeddings=NUM_PV_SYSTEMS + NUM_GSPS,
            embedding_dim=self.pv_system_id_embedding_dim,
        )

        self.pv_query_generator = PVQueryGenerator(
            pv_system_id_embedding=id_embedding,
            num_gsps=NUM_GSPS,
        )

        self.gsp_query_generator = GSPQueryGenerator(gsp_id_embedding=id_embedding)

        self.transformer_encoder = MultiLayerTransformerEncoder(
            d_model=self.d_model,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> dict[str, torch.Tensor]:
        # Reshape so each timestep is considered a different example:
        num_examples, num_timesteps = x[BatchKey.pv].shape[:2]

        # Process satellite data and queries:
        pv_query = self.pv_query_generator(x)  # shape: example * time, n_pv_systems, features
        gsp_query = self.gsp_query_generator(
            x,
            include_history=True,
            base_batch_key=BatchKey.gsp_5_min,
            do_reshape_time_as_batch=True,
        )  # shape: example * time, 1, features
        satellite_data = self.hrvsatellite_processor(x)  # shape: example * time, positions, feats.

        # Pad with zeros if necessary to get up to self.d_model:
        pv_query = maybe_pad_with_zeros(pv_query, requested_dim=self.d_model)
        gsp_query = maybe_pad_with_zeros(gsp_query, requested_dim=self.d_model)
        satellite_data = maybe_pad_with_zeros(satellite_data, requested_dim=self.d_model)

        # Prepare inputs for the transformer_encoder:
        attn_input = torch.concat((pv_query, gsp_query, satellite_data), dim=1)

        # Mask the NaN GSP and PV queries. True or non-zero value indicates value will be ignored.
        mask = attn_input.isnan().any(dim=2)
        attn_input = attn_input.nan_to_num(0)

        # Pass data into transformer_encoder:
        attn_output = attn_input + self.transformer_encoder(attn_input, src_key_padding_mask=mask)

        # Reshape to (example time element d_model):
        attn_output = einops.rearrange(
            attn_output,
            "(example time) ... -> example time ...",
            example=num_examples,
            time=num_timesteps,
        )

        assert attn_output.isfinite().all()

        # Select the elements of the output which correspond to the query:
        gsp_start_idx = pv_query.shape[1]
        gsp_end_idx = gsp_start_idx + gsp_query.shape[1]
        pv_attn_out = attn_output[:, :, :gsp_start_idx]
        gsp_attn_out = attn_output[:, :, gsp_start_idx:gsp_end_idx]

        return {
            "pv_attn_out": pv_attn_out,  # shape: (example, 5_min_time, n_pv_systems, d_model)
            "gsp_attn_out": gsp_attn_out,  # shape: (example, 5_min_time, 1, d_model)
        }


@dataclass(eq=False)
class PVRNN(nn.Module):
    hidden_size: int
    num_layers: int
    dropout: float

    def __post_init__(self):
        super().__init__()

        rnn_kwargs = dict(
            hidden_size=self.hidden_size,
            num_layers=self.num_layers,
            dropout=self.dropout,
            batch_first=True,
        )
        self.pv_rnn_history_encoder = nn.RNN(
            # Each timestep of the encoder RNN receives the output of the satellite_transformer
            # (which is of size d_model per PV system and per timestep), for a single PV system,
            # plus one timestep of that PV system's history.
            input_size=self.hidden_size + 1,
            **rnn_kwargs,
        )
        self.pv_rnn_future_decoder = nn.RNN(
            # Each timestep of the decoder RNN receives the output of the satellite_transformer
            # (which is of size d_model per PV system and per timestep) for a single PV system.
            input_size=self.hidden_size,
            **rnn_kwargs,
        )

    def forward(
        self, x: dict[BatchKey, torch.Tensor], sat_trans_pv_attn_out: torch.Tensor
    ) -> torch.Tensor:
        # Prepare inputs for the pv_rnn_history_encoder:
        # Each timestep of the encoder RNN receives the output of the satellite_transformer
        # (which is of size d_model per PV system and per timestep), for a single PV system,
        # plus one timestep of that PV system's history.
        # Some of `x[BatchKey.pv]` will be NaN. But that's OK. We mask missing examples.
        # And use nan_to_num later in this function.
        pv_t0_idx = x[BatchKey.pv_t0_idx]
        num_pv_timesteps = x[BatchKey.pv].shape[1]
        hist_pv = x[BatchKey.pv][:, : pv_t0_idx + 1]  # Shape: example, time, n_pv_systems

        # Reshape so each PV system is seen as a different example:
        hist_pv = einops.rearrange(
            hist_pv, "example time n_pv_systems -> (example n_pv_systems) time 1"
        )
        sat_trans_pv_attn_out = einops.rearrange(
            sat_trans_pv_attn_out,
            "example time n_pv_systems d_model -> (example n_pv_systems) time d_model",
            n_pv_systems=N_PV_SYSTEMS_PER_EXAMPLE,  # sanity check
            time=num_pv_timesteps,
        )
        hist_sat_trans_pv_attn_out = sat_trans_pv_attn_out[:, : pv_t0_idx + 1]

        pv_rnn_hist_enc_in = torch.concat((hist_pv, hist_sat_trans_pv_attn_out), dim=2)
        pv_rnn_hist_enc_in = pv_rnn_hist_enc_in.nan_to_num(0)
        pv_rnn_hist_enc_out, pv_rnn_hist_enc_hidden = self.pv_rnn_history_encoder(
            pv_rnn_hist_enc_in
        )

        # Now for the pv_rnn_future_decoder:
        future_sat_trans_pv_attn_out = sat_trans_pv_attn_out[:, pv_t0_idx + 1 :]
        future_sat_trans_pv_attn_out = future_sat_trans_pv_attn_out.nan_to_num(0)
        pv_rnn_fut_dec_out, _ = self.pv_rnn_future_decoder(
            future_sat_trans_pv_attn_out,
            pv_rnn_hist_enc_hidden,
        )

        # Concatenate the output from the encoder and decoder RNNs, and reshape
        # so each timestep and each PV system is a separate element into `time_transformer`.
        pv_rnn_out = torch.concat((pv_rnn_hist_enc_out, pv_rnn_fut_dec_out), dim=1)
        # rnn_out shape: (example n_pv_systems), time, d_model
        assert pv_rnn_out.isfinite().all()
        pv_rnn_out = einops.rearrange(
            pv_rnn_out,
            "(example n_pv_systems) time d_model -> example (time n_pv_systems) d_model",
            n_pv_systems=N_PV_SYSTEMS_PER_EXAMPLE,
            d_model=self.hidden_size,
            time=num_pv_timesteps,
        )
        return pv_rnn_out


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class FullModel(pl.LightningModule):
    d_model: int = D_MODEL
    num_heads: int = N_HEADS
    dropout: float = 0.1
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 8
    cheat: bool = False  #: Use real satellite imagery of the future.
    #: Compute the loss on a central crop of the imagery.
    num_5_min_history_timesteps_during_training: Optional[int] = 4
    num_5_min_forecast_timesteps_during_training: Optional[int] = 6
    num_gaussians: int = 4
    num_rnn_layers: int = 4

    def __post_init__(self):
        super().__init__()

        if self.cheat:
            _log.warning("CHEATING MODE ENABLED! Using real satellite imagery of future!")

        # Predict future satellite images:
        self.satellite_predictor = SatellitePredictor()

        # Load SatellitePredictor weights
        self.satellite_predictor.load_state_dict(
            torch.load(
                (
                    "/home/jack/dev/ocf/power_perceiver/power_perceiver/experiments/"
                    "power_perceiver/3qvkf1dy/checkpoints/"
                    "epoch=170-step=175104-just-satellite-predictor.state_dict.pth"
                )
                if ON_DONATELLO
                else (
                    "/home/jack/model_params/satellite_predictor/"
                    "epoch=170-step=175104-just-satellite-predictor.state_dict.pth"
                )
            )
        )

        # Infer GSP and PV power output for a single timestep of satellite imagery.
        self.satellite_transformer = SatelliteTransformer()

        self.nwp_processor = NWPProcessor(n_channels=len(NWP_CHANNELS))

        # Find temporal features, and help calibrate predictions using recent history:
        self.pv_rnn = PVRNN(
            hidden_size=self.d_model,
            num_layers=self.num_rnn_layers,
            dropout=self.dropout,
        )

        # Look across timesteps to produce the final output.
        self.time_transformer = MultiLayerTransformerEncoder(
            d_model=self.d_model,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

        self.pv_mixture_density_net = nn.Sequential(
            nn.Linear(in_features=self.d_model, out_features=self.d_model),
            nn.GELU(),
            nn.Linear(in_features=self.d_model, out_features=self.d_model),
            nn.GELU(),
            MixtureDensityNetwork(in_features=self.d_model, num_gaussians=self.num_gaussians),
        )

        # Do this at the end of __post_init__ to capture model topology to wandb:
        self.save_hyperparameters()

    def _predict_satellite(self, x: dict[BatchKey, torch.Tensor]) -> dict[BatchKey, torch.Tensor]:
        """Predict future satellite images. The SatellitePredictor always gets every timestep."""
        if self.cheat:
            predicted = x[BatchKey.hrvsatellite_actual][:, :NUM_HIST_SAT_IMAGES, 0]
        else:
            predicted = self.satellite_predictor(x=x)  # Shape: example, time, y, x

        assert predicted.isfinite().all()
        x[BatchKey.hrvsatellite_predicted] = predicted
        return x

    def _select_random_subset_of_timesteps(
        self, x: dict[BatchKey, torch.Tensor]
    ) -> dict[BatchKey, torch.Tensor]:
        random_history_timestep_indexes = random_int_without_replacement(
            start=0,
            stop=NUM_HIST_SAT_IMAGES,
            num=self.num_5_min_history_timesteps_during_training,
        )
        random_forecast_timestep_indexes = random_int_without_replacement(
            start=NUM_HIST_SAT_IMAGES,
            stop=NUM_HIST_SAT_IMAGES + NUM_FUTURE_SAT_IMAGES,
            num=self.num_5_min_forecast_timesteps_during_training,
        )
        random_timestep_indexes = np.concatenate(
            (random_history_timestep_indexes, random_forecast_timestep_indexes)
        )
        x[BatchKey.hrvsatellite_actual] = x[BatchKey.hrvsatellite_actual][
            :, random_history_timestep_indexes
        ]
        x[BatchKey.hrvsatellite_predicted] = x[BatchKey.hrvsatellite_predicted][
            :, random_forecast_timestep_indexes - NUM_HIST_SAT_IMAGES
        ]
        for batch_key in (
            # If we go back to training the u-net, then we'll have to save `hrvsatellite`
            # before it is subsampled. See issue #156.
            # Don't include BatchKey.hrvsatellite_time_utc because all it's used for
            # is plotting satellite predictions, and we need all timesteps for that!
            # Subselect pv_time_utc here, so `plot_probabability_timeseries.plot_pv_power`
            # works correctly.
            # We *do* subset hrvsatellite_time_utc_fourier because it's used in the
            # satellite_transformer.
            BatchKey.hrvsatellite_time_utc_fourier,
            BatchKey.hrvsatellite_solar_azimuth,
            BatchKey.hrvsatellite_solar_elevation,
            BatchKey.pv,
            BatchKey.pv_time_utc,
            BatchKey.pv_time_utc_fourier,
            BatchKey.pv_solar_azimuth,
            BatchKey.pv_solar_elevation,
            BatchKey.gsp_5_min,
            BatchKey.gsp_5_min_time_utc_fourier,
            BatchKey.gsp_5_min_solar_azimuth,
            BatchKey.gsp_5_min_solar_elevation,
        ):
            x[batch_key] = x[batch_key][:, random_timestep_indexes]

        for batch_key in (
            BatchKey.hrvsatellite_t0_idx,
            BatchKey.pv_t0_idx,
            BatchKey.gsp_5_min_t0_idx,
        ):
            x[batch_key] = self.num_5_min_history_timesteps_during_training - 1

        return x

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> dict[str, torch.Tensor]:
        # Predict future satellite images. The SatellitePredictor always gets every timestep.
        x = self._predict_satellite(x)

        # Select a subset of 5-minute timesteps during training:
        if self.num_5_min_history_timesteps_during_training and self.training:
            x = self._select_random_subset_of_timesteps(x)
            num_5_min_timesteps = (
                self.num_5_min_history_timesteps_during_training
                + self.num_5_min_forecast_timesteps_during_training
            )
        else:
            num_5_min_timesteps = NUM_HIST_SAT_IMAGES + NUM_FUTURE_SAT_IMAGES

        x = _crop_satellite_and_spatial_coords(x)

        # Pass through the transformer!
        sat_trans_out = self.satellite_transformer(x=x)

        # Pass through PV RNN:
        sat_trans_pv_attn_out = sat_trans_out["pv_attn_out"]
        # Shape of `sat_trans_pv_attn_out`: (example time n_pv_systems d_model)
        pv_rnn_out = self.pv_rnn(x=x, sat_trans_pv_attn_out=sat_trans_pv_attn_out)
        del sat_trans_pv_attn_out
        n_pv_elements = pv_rnn_out.shape[1]

        # ----------------- Prepare inputs for the `time_transformer` -----------------------------

        # Get GSP attention outputs. Each GSP attention output timestep is seen as a separate
        # element into the `time transformer`. Remember, at this stage, `sat_trans_gsp_attn_out`
        # is 5-minutely. Shape (after squeeze): example time d_model.
        sat_trans_gsp_attn_out = sat_trans_out["gsp_attn_out"].squeeze()

        # Pad with zeros
        pv_rnn_out = maybe_pad_with_zeros(pv_rnn_out, requested_dim=self.d_model)
        sat_trans_gsp_attn_out = maybe_pad_with_zeros(
            sat_trans_gsp_attn_out, requested_dim=self.d_model
        )

        # Get GSP query for the time_transformer:
        # The query for the time_transformer is just for the half-hourly timesteps
        # (not the 5-minutely GSP queries used in the `SatelliteTransformer`.)
        gsp_query_generator: GSPQueryGenerator = self.satellite_transformer.gsp_query_generator
        gsp_query = gsp_query_generator.forward(
            x,
            include_history=True,
            base_batch_key=BatchKey.gsp,
            do_reshape_time_as_batch=False,
        )
        gsp_query = maybe_pad_with_zeros(gsp_query, requested_dim=self.d_model)

        # Prepare NWP inputs
        nwp_query = self.nwp_processor(x)
        nwp_query = maybe_pad_with_zeros(nwp_query, requested_dim=self.d_model)

        # Concatenate all the things we're going to feed into the "time transformer":
        # Shape: (example, elements, d_model)
        # `pv_rnn_out` must be the first set of elements.
        # `gsp_query` must be the last set of elements.
        # The shape is (example, query_elements, d_model).
        time_attn_in = torch.concat(
            (pv_rnn_out, sat_trans_gsp_attn_out, nwp_query, gsp_query), dim=1
        )

        # It's necessary to mask the time_transformer because some examples, where we're trying to
        # predict GSP PV power, might have NaN PV data (because the GSP is so far north it has no
        # PV systems!), and we don't want the attention mechanism to pay any attention to the
        # zero'd out (previously NaN) PV data when predicting GSP PV power.
        mask = time_attn_in.isnan().any(dim=2)
        time_attn_in = time_attn_in.nan_to_num(0)
        time_attn_out = self.time_transformer(time_attn_in, src_key_padding_mask=mask)

        # ---------------------------- MIXTURE DENSITY NETWORK -----------------------------------
        # The MDN doesn't like NaNs, and I think there will be NaNs when, for example,
        # the example is missing PV data:
        time_attn_out = time_attn_out.nan_to_num(0)
        predicted_pv_power = self.pv_mixture_density_net(time_attn_out[:, :n_pv_elements])

        # Reshape the PV power predictions
        predicted_pv_power = einops.rearrange(
            predicted_pv_power,
            "example (time n_pv_systems) mdn_features -> example time n_pv_systems mdn_features",
            time=num_5_min_timesteps,
            n_pv_systems=N_PV_SYSTEMS_PER_EXAMPLE,
            mdn_features=self.num_gaussians * 3,  # x3 for pi, mu, sigma.
        )

        # GSP power. There's just 1 GSP. So each gsp element is a timestep.
        n_gsp_elements = gsp_query.shape[1]
        predicted_gsp_power = self.pv_mixture_density_net(time_attn_out[:, -n_gsp_elements:])

        return dict(
            predicted_pv_power=predicted_pv_power,  # Shape: (example time n_pv_sys mdn_features)
            predicted_gsp_power=predicted_gsp_power,  # Shape: (example time mdn_features)
            predicted_sat=x[BatchKey.hrvsatellite_predicted],  # Shape: example, time, y, x
            pv_time_utc=x[BatchKey.pv_time_utc],  # Give the subsampled times to plotting functions.
        )

    @property
    def tag(self) -> str:
        return "train" if self.training else "validation"

    def training_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        return self.validation_step(batch, batch_idx)

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:

        network_out = self(batch)

        # PV & GSP LOSS ######################
        predicted_pv_power = network_out["predicted_pv_power"]
        actual_pv_power = batch[BatchKey.pv]
        predicted_gsp_power = network_out["predicted_gsp_power"]
        actual_gsp_power = batch[BatchKey.gsp].squeeze()

        # Mask predicted and actual PV and GSP (some examples don't have PV and/or GSP)
        # For more discussion of how to mask losses in pytorch, see:
        # https://discuss.pytorch.org/t/masking-input-to-loss-function/121830/3
        pv_mask = actual_pv_power.isfinite()
        pv_t0_idx = batch[BatchKey.pv_t0_idx]
        pv_slice_from_t0 = slice(pv_t0_idx + 1, None)
        pv_mask_from_t0 = pv_mask[:, pv_slice_from_t0]

        gsp_mask = actual_gsp_power.isfinite()
        gsp_t0_idx = batch[BatchKey.gsp_t0_idx]
        gsp_slice_from_t0 = slice(gsp_t0_idx + 1, None)
        gsp_slice_from_t0_to_1h = slice(gsp_t0_idx + 1, gsp_t0_idx + 3)
        gsp_mask_from_t0 = gsp_mask[:, gsp_slice_from_t0]
        gsp_mask_from_t0_to_1h = gsp_mask[:, gsp_slice_from_t0_to_1h]

        predicted_pv_power_from_t0 = predicted_pv_power[:, pv_slice_from_t0][pv_mask_from_t0]
        actual_pv_power_from_t0 = actual_pv_power[:, pv_slice_from_t0][pv_mask_from_t0]

        predicted_gsp_power_from_t0 = predicted_gsp_power[:, gsp_slice_from_t0][gsp_mask_from_t0]
        actual_gsp_power_from_t0 = actual_gsp_power[:, gsp_slice_from_t0][gsp_mask_from_t0]

        predicted_gsp_power_from_t0_to_1h = predicted_gsp_power[:, gsp_slice_from_t0_to_1h][
            gsp_mask_from_t0_to_1h
        ]
        actual_gsp_power_from_t0_to_1h = actual_gsp_power[:, gsp_slice_from_t0_to_1h][
            gsp_mask_from_t0_to_1h
        ]

        # PV negative log prob loss:
        pv_distribution = get_distribution(predicted_pv_power_from_t0)
        pv_neg_log_prob_loss = -pv_distribution.log_prob(actual_pv_power_from_t0).mean()
        self.log(f"{self.tag}/pv_neg_log_prob", pv_neg_log_prob_loss)

        # PV power loss:
        pv_mse_loss = F.mse_loss(pv_distribution.mean, actual_pv_power_from_t0)
        self.log(f"{self.tag}/pv_mse", pv_mse_loss)

        pv_nmae_loss = F.l1_loss(pv_distribution.mean, actual_pv_power_from_t0)
        self.log(f"{self.tag}/pv_nmae", pv_nmae_loss)

        # GSP negative log prob loss from t0:
        gsp_distribution_from_t0 = get_distribution(predicted_gsp_power_from_t0)
        gsp_neg_log_prob_loss_from_t0 = -gsp_distribution_from_t0.log_prob(
            actual_gsp_power_from_t0
        ).mean()
        self.log(f"{self.tag}/gsp_neg_log_prob", gsp_neg_log_prob_loss_from_t0)

        # GSP negative log prob loss from t0 to 1h forecast:
        gsp_distribution_from_t0_to_1h = get_distribution(predicted_gsp_power_from_t0_to_1h)
        gsp_neg_log_prob_loss_from_t0_to_1h = -gsp_distribution_from_t0_to_1h.log_prob(
            actual_gsp_power_from_t0_to_1h
        ).mean()
        self.log(f"{self.tag}/gsp_neg_log_prob_from_t0_to_1h", gsp_neg_log_prob_loss_from_t0_to_1h)

        # GSP power loss:
        gsp_mse_loss = F.mse_loss(gsp_distribution_from_t0.mean, actual_gsp_power_from_t0)
        self.log(f"{self.tag}/gsp_mse", gsp_mse_loss)

        gsp_nmae_loss = F.l1_loss(gsp_distribution_from_t0.mean, actual_gsp_power_from_t0)
        self.log(f"{self.tag}/gsp_nmae", gsp_nmae_loss)

        # Total PV and GSP MSE loss:
        total_pv_and_gsp_mse_loss = gsp_mse_loss + pv_mse_loss
        self.log(f"{self.tag}/total_mse", total_pv_and_gsp_mse_loss)

        # Total NMAE loss:
        total_nmae_loss = pv_nmae_loss + gsp_nmae_loss
        self.log(f"{self.tag}/total_nmae", total_nmae_loss)

        # Total PV and GSP negative log probability loss:
        total_pv_and_gsp_neg_log_prob_loss = (
            gsp_neg_log_prob_loss_from_t0
            + (gsp_neg_log_prob_loss_from_t0_to_1h * 2)
            + pv_neg_log_prob_loss
        )
        self.log(
            f"{self.tag}/total_pv_and_gsp_neg_log_prob_loss", total_pv_and_gsp_neg_log_prob_loss
        )

        return {
            "loss": total_pv_and_gsp_neg_log_prob_loss,
            "predicted_gsp_power": predicted_gsp_power,
            "predicted_gsp_power_mean": get_distribution(predicted_gsp_power).mean,
            "actual_gsp_power": actual_gsp_power,
            "gsp_time_utc": batch[BatchKey.gsp_time_utc],
            "actual_pv_power": actual_pv_power,
            "predicted_pv_power": predicted_pv_power,
            "predicted_pv_power_mean": get_distribution(predicted_pv_power).mean,
            "pv_time_utc": network_out["pv_time_utc"],
            "predicted_sat": network_out["predicted_sat"],  # Shape: example, time, y, x
            "actual_sat": batch[BatchKey.hrvsatellite_actual][:, NUM_HIST_SAT_IMAGES:, 0],
        }

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=5e-5)

        def _lr_lambda(epoch):
            return 50 / (epoch + 50)

        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, _lr_lambda, verbose=True)
        return [optimizer], [scheduler]


# ---------------------------------- Training ----------------------------------

if __name__ == "__main__":
    # Important to seed the models when using DistributedDataProcessing, so the
    # models on different GPUs are initialised the same way. But we *do* want
    # to seed our data loader workers differently for each GPU, so we do that
    # in our own worker_init_fn.
    pl.seed_everything(42)

    if DEBUG:
        train_time_periods = pd.DataFrame({"start_dt": "2020-01-01", "end_dt": "2020-03-01"})
        val_time_periods = pd.DataFrame({"start_dt": "2021-01-01", "end_dt": "2021-03-01"})
    else:
        # Train on all of 2020. Plus every 5th week of 2021.
        every_week = pd.period_range("2020-01-01", "2021-12-31", freq="W")
        val_weeks = every_week[53::5]
        train_weeks = set(every_week) - set(val_weeks)
        train_weeks = pd.PeriodIndex(train_weeks).sort_values()

        # Now convert to our "time_periods" DataFrame:
        train_time_periods = pandas_periods_to_our_periods_dt(train_weeks)
        val_time_periods = pandas_periods_to_our_periods_dt(val_weeks)

    # Get data loaders
    train_dataloader = get_dataloader(
        time_periods=train_time_periods,
        num_workers=0 if DEBUG else 4,
        n_batches_per_epoch_per_worker=64 if DEBUG else 1024,
        load_subset_every_epoch=True,
        train=True,
    )

    N_GSPS_AFTER_FILTERING = 313
    val_dataloader = get_dataloader(
        time_periods=val_time_periods,
        # num_workers for NationalPVDataset MUST BE SAME 1!
        # OTHERWISE LogNationalPV BREAKS! See:
        # https://github.com/openclimatefix/power_perceiver/issues/130
        num_workers=0 if DEBUG else 1,
        n_batches_per_epoch_per_worker=64 if DEBUG else N_GSPS_AFTER_FILTERING,
        load_subset_every_epoch=False,
        train=False,
    )

    # Init model:
    model = FullModel()

    if ENABLE_WANDB:
        wandb_logger = WandbLogger(
            name="027.08: Train on more data. GCP-1",
            project="power_perceiver",
            entity="openclimatefix",
            log_model=True,
        )
        callbacks = [
            # Save the top 3 model params
            pl.callbacks.ModelCheckpoint(
                monitor="validation/gsp_nmae",
                mode="min",
                save_top_k=3,
            ),
            # Always save the most recent model (so we can resume training)
            pl.callbacks.ModelCheckpoint(filename="{epoch}"),
            pl.callbacks.LearningRateMonitor(logging_interval="step"),
            LogProbabilityTimeseriesPlots(),
            LogTSNEPlot(query_generator_name="satellite_transformer.pv_query_generator"),
            LogNationalPV(),
        ]
    else:
        wandb_logger = False
        callbacks = None

    # WARNING: Don't run multiple GPUs in ipython.
    trainer = pl.Trainer(
        gpus=GPUS,
        strategy="ddp" if len(GPUS) > 1 else None,
        max_epochs=200,
        logger=wandb_logger,
        callbacks=callbacks,
    )

    trainer.fit(
        model=model,
        train_dataloaders=train_dataloader,
        val_dataloaders=val_dataloader,
    )



# General imports
import logging
import socket
from dataclasses import dataclass, field
from pathlib import Path

import einops

# ML imports
import matplotlib.pyplot as plt
import pandas as pd
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from pytorch_lightning.loggers import WandbLogger
from pytorch_msssim import ms_ssim

from power_perceiver.analysis.plot_satellite import LogSatellitePlots

# power_perceiver imports
from power_perceiver.consts import X_OSGB_MEAN, X_OSGB_STD, Y_OSGB_MEAN, Y_OSGB_STD, BatchKey
from power_perceiver.load_prepared_batches.data_sources import HRVSatellite
from power_perceiver.load_prepared_batches.data_sources.satellite import SAT_MEAN, SAT_STD
from power_perceiver.load_prepared_batches.prepared_dataset import PreparedDataset
from power_perceiver.load_raw.data_sources.satellite_zarr_dataset import (
    SatelliteZarrDataset,
    worker_init_fn,
)
from power_perceiver.np_batch_processor.sun_position import SunPosition
from power_perceiver.np_batch_processor.topography import Topography
from power_perceiver.pytorch_modules.satellite_predictor import XResUNet

logging.basicConfig()
_log = logging.getLogger("power_perceiver")
_log.setLevel(logging.DEBUG)

plt.rcParams["figure.figsize"] = (18, 10)
plt.rcParams["figure.facecolor"] = "white"

NUM_HIST_SAT_IMAGES = 7  # v15 pre-prepared batches use 7
NUM_FUTURE_SAT_IMAGES = 24  # v15 pre-prepared batches use 24
IMAGE_SIZE_PIXELS = 64  # v15 pre-prepared batches use 64
USE_TOPOGRAPHY = True
USE_SUN_POSITION = True

if socket.gethostname() == "donatello":
    SATELLITE_ZARR_PATH = (
        "/mnt/storage_ssd_4tb/data/ocf/solar_pv_nowcasting/nowcasting_dataset_pipeline/"
        "satellite/EUMETSAT/SEVIRI_RSS/zarr/v3/eumetsat_seviri_hrv_uk.zarr"
    )
    DATA_PATH = Path(
        "/mnt/storage_ssd_4tb/data/ocf/solar_pv_nowcasting/nowcasting_dataset_pipeline/"
        "prepared_ML_training_data/v15"
    )
else:
    # On Google Cloud VM:
    SATELLITE_ZARR_PATH = (
        "gs://solar-pv-nowcasting-data/satellite/EUMETSAT/SEVIRI_RSS/v3/eumetsat_seviri_hrv_uk.zarr"
    )
    DATA_PATH = Path("/home/jack/data/v15")

assert DATA_PATH.exists()


torch.manual_seed(42)

np_batch_processors = []
if USE_SUN_POSITION:
    np_batch_processors.append(SunPosition(t0_timestep_idx=NUM_HIST_SAT_IMAGES - 1))
if USE_TOPOGRAPHY:
    np_batch_processors.append(Topography("/home/jack/europe_dem_2km_osgb.tif"))


def get_satellite_zarr_dataloader(**kwargs):
    return torch.utils.data.DataLoader(
        SatelliteZarrDataset(
            satellite_zarr_path=SATELLITE_ZARR_PATH,
            np_batch_processors=np_batch_processors,
            size_pixels=IMAGE_SIZE_PIXELS,
            n_timesteps_per_example=NUM_HIST_SAT_IMAGES + NUM_FUTURE_SAT_IMAGES,
            **kwargs,
        ),
        batch_size=32,
        num_workers=4,
        pin_memory=True,
        worker_init_fn=worker_init_fn,
        persistent_workers=True,
    )


train_dataloader = get_satellite_zarr_dataloader()


if IMAGE_SIZE_PIXELS == 64 and NUM_HIST_SAT_IMAGES == 7:
    # Use pre-prepared batches:
    _log.info("Using pre-prepared batches for validation set.")
    val_dataloader = torch.utils.data.DataLoader(
        PreparedDataset(
            data_path=DATA_PATH / "test",
            data_loaders=[
                HRVSatellite(),
            ],
            np_batch_processors=np_batch_processors,
        ),
        batch_size=None,
        num_workers=12,
        pin_memory=True,
        persistent_workers=True,
    )
else:
    _log.info("Using SatelliteZarrDataset for validation set.")
    val_dataloader = get_satellite_zarr_dataloader(
        start_date=pd.Timestamp("2021-01-01 00:00"),
        end_date=pd.Timestamp("2021-12-31 23:59"),
        load_once=True,
        n_days_to_load_per_epoch=2,  # Don't use up too much RAM!
    )


def get_osgb_coords_for_coord_conv(batch: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
    """Returns tensor of shape (example, 2, y, x)."""
    y_osgb = batch[BatchKey.hrvsatellite_y_osgb]
    x_osgb = batch[BatchKey.hrvsatellite_x_osgb]

    # Normalise:
    y_osgb = (y_osgb - Y_OSGB_MEAN) / Y_OSGB_STD
    x_osgb = (x_osgb - X_OSGB_MEAN) / X_OSGB_STD

    # Concat:
    return torch.stack((y_osgb, x_osgb), dim=1)


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class SatellitePredictor(pl.LightningModule):
    use_coord_conv: bool = True
    crop: bool = False
    optimizer_class: torch.optim.Optimizer = torch.optim.Adam
    optimizer_kwargs: dict = field(
        # lambda trick from https://stackoverflow.com/a/52064202/732596
        default_factory=lambda: dict(lr=1e-4)
    )
    use_topography: bool = USE_TOPOGRAPHY
    use_sun_position: bool = USE_SUN_POSITION

    # kwargs to fastai DynamicUnet. See this page for details:
    # https://fastai1.fast.ai/vision.models.unet.html#DynamicUnet
    pretrained: bool = False
    blur_final: bool = True  # Blur final layer. fastai default is True.
    self_attention: bool = True  # Use SA layer at the third block before the end.
    last_cross: bool = True  # Use a cross-connection with the direct input of the model.
    bottle: bool = False  # Bottleneck the last skip connection.

    def __post_init__(self):
        super().__init__()

        self.satellite_predictor = XResUNet(
            img_size=(IMAGE_SIZE_PIXELS, IMAGE_SIZE_PIXELS),
            n_in=(
                NUM_HIST_SAT_IMAGES
                + (2 if self.use_coord_conv else 0)
                + (1 if self.use_topography else 0)
                + (2 if self.use_sun_position else 0)
            ),
            n_out=NUM_FUTURE_SAT_IMAGES,
            pretrained=self.pretrained,
            blur_final=self.blur_final,
            self_attention=self.self_attention,
            last_cross=self.last_cross,
            bottle=self.bottle,
        )

        # Do this at the end of __post_init__ to capture model topology to wandb:
        self.save_hyperparameters()

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> dict[str, torch.Tensor]:
        data = x[BatchKey.hrvsatellite_actual][
            :, :NUM_HIST_SAT_IMAGES, 0
        ]  # Shape: (example, time, y, x)
        height, width = data.shape[2:]
        if self.use_coord_conv:
            osgb_coords = get_osgb_coords_for_coord_conv(x)
            data = torch.concat((data, osgb_coords), dim=1)

        if self.use_topography:
            surface_height = x[BatchKey.hrvsatellite_surface_height]
            surface_height = surface_height.unsqueeze(1)  # Add channel dim
            data = torch.concat((data, surface_height), dim=1)

        if self.use_sun_position:
            azimuth = x[BatchKey.solar_azimuth_at_t0]
            elevation = x[BatchKey.solar_elevation_at_t0]
            sun_pos = torch.stack((azimuth, elevation), dim=1)  # Shape: (example, 2)
            del azimuth, elevation
            # Repeat over y and x:
            sun_pos = einops.repeat(sun_pos, "example chan -> example chan y x", y=height, x=width)
            data = torch.concat((data, sun_pos), dim=1)

        predicted_sat = self.satellite_predictor(data)
        return dict(predicted_sat=predicted_sat)

    def training_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        return self.validation_step(batch, batch_idx)

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        tag = "train" if self.training else "validation"
        network_out = self(batch)
        predicted_sat = network_out["predicted_sat"]
        actual_sat = batch[BatchKey.hrvsatellite_actual][:, NUM_HIST_SAT_IMAGES:, 0]
        sat_mse_loss = F.mse_loss(predicted_sat, actual_sat)
        self.log(f"{tag}/sat_mse", sat_mse_loss)

        # MS-SSIM. Requires images to be de-normalised:
        actual_sat_denorm = (actual_sat * SAT_STD["HRV"]) + SAT_MEAN["HRV"]
        predicted_sat_denorm = (predicted_sat * SAT_STD["HRV"]) + SAT_MEAN["HRV"]
        ms_ssim_loss = 1 - ms_ssim(
            predicted_sat_denorm,
            actual_sat_denorm,
            data_range=1023.0,
            size_average=True,  # Return a scalar.
            win_size=3,  # ClimateHack folks used win_size=3.
        )
        self.log(f"{tag}/ms_ssim", ms_ssim_loss)
        self.log(f"{tag}/ms_ssim+sat_mse", ms_ssim_loss + sat_mse_loss)

        if self.crop:
            # Loss on 33x33 central crop:
            # The image has to be larger than 32x32 otherwise ms-ssim complains:
            # "Image size should be larger than 32 due to the 4 downsamplings in ms-ssim"
            CROP = 15
            sat_mse_loss_crop = F.mse_loss(
                predicted_sat[:, :, CROP:-CROP, CROP:-CROP],
                actual_sat[:, :, CROP:-CROP, CROP:-CROP],
            )
            self.log(f"{tag}/sat_mse_crop", sat_mse_loss_crop)
            ms_ssim_loss_crop = 1 - ms_ssim(
                predicted_sat_denorm[:, :, CROP:-CROP, CROP:-CROP],
                actual_sat_denorm[:, :, CROP:-CROP, CROP:-CROP],
                data_range=1023.0,
                size_average=True,  # Return a scalar.
                win_size=3,  # ClimateHack folks used win_size=3.
            )
            self.log(f"{tag}/ms_ssim_crop", ms_ssim_loss_crop)
            self.log(f"{tag}/ms_ssim_crop+sat_mse_crop", ms_ssim_loss_crop + sat_mse_loss_crop)
            loss = ms_ssim_loss_crop + sat_mse_loss_crop
        else:
            loss = ms_ssim_loss + sat_mse_loss

        return dict(
            loss=loss,
            predicted_sat=predicted_sat,
            actual_sat=actual_sat,
        )

    def configure_optimizers(self):
        optimizer = self.optimizer_class(self.parameters(), **self.optimizer_kwargs)

        def _lr_lambda(epoch):
            return 50 / (epoch + 50)

        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, _lr_lambda, verbose=True)
        return [optimizer], [scheduler]


model = SatellitePredictor()

wandb_logger = WandbLogger(
    name=(
        "022.27: Coord conv. Load 64 days. Use all data where Sun >= 5 degrees above horizon."
        " donatello-4"
    ),
    project="power_perceiver",
    entity="openclimatefix",
    log_model="all",
)

# log model only if validation loss decreases
if model.crop:
    loss_name = "validation/ms_ssim_crop+sat_mse_crop"
else:
    loss_name = "validation/ms_ssim+sat_mse"
checkpoint_callback = pl.callbacks.ModelCheckpoint(monitor=loss_name, mode="min")


if socket.gethostname() == "donatello":
    GPU = 4
else:  # On GCP
    GPU = 0
trainer = pl.Trainer(
    gpus=[GPU],
    max_epochs=200,
    logger=wandb_logger,
    callbacks=[
        LogSatellitePlots(),
        checkpoint_callback,
        pl.callbacks.LearningRateMonitor(logging_interval="step"),
    ],
)

trainer.fit(
    model=model,
    train_dataloaders=train_dataloader,
    val_dataloaders=val_dataloader,
)
